import base64
from typing import TypeVar

import regex as re

from ..core import ChepyCore, ChepyDecorators

LinksT = TypeVar("LinksT", bound="Links")


class Links(ChepyCore):
    def __init__(self, *data):
        super().__init__(*data)

    @ChepyDecorators.call_stack
    def pastebin_to_raw(self) -> LinksT:
        """Convert a pastebin link to raw pastebin link

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("https://pastebin.com/abCD").pastebin_to_raw()
            'https://pastebin.com/raw/abCD'
        """
        self.state = re.sub(r"(pastebin\.com)(/)", r"\1/raw\2", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def github_to_raw(self) -> LinksT:
        """Convert a github link to raw github link

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("https://github.com/securisec/chepy/blob/master/README.md").github_to_raw()
            'https://raw.githubusercontent.com/securisec/chepy/master/README.md'
        """
        self.state = re.sub(
            "/blob",
            "",
            re.sub(
                r"(github\.com)(/)",
                r"raw.githubusercontent.com\2",
                self._convert_to_str(),
            ),
        )
        return self

    @ChepyDecorators.call_stack
    def google_search_ei_to_epoch(self) -> LinksT:
        """Convert a google search ei query param to epoch

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("Bh8hYqykHc64mAXkkoTgCg==").google_search_ei_to_epoch()
            1646337798
        """
        decoded = base64.urlsafe_b64decode(self._convert_to_str())
        timestamp = ord(chr(decoded[0]))
        timestamp += ord(chr(decoded[1])) * 256
        timestamp += ord(chr(decoded[2])) * 256**2
        timestamp += ord(chr(decoded[3])) * 256**3
        self.state = timestamp
        return self



from ..core import ChepyCore
from typing import Any, TypeVar, Literal, Dict, Union

jwt: Any
AES: Any
ARC4: Any
DES: Any
DES3: Any
Blowfish: Any
EncryptionEncodingT = TypeVar('EncryptionEncodingT', bound='EncryptionEncoding')
FORMAT = Literal['hex', 'base64', 'utf-8', 'latin-1', 'raw']
RC4_FORMAT = Literal['hex', 'base64', 'utf8', 'utf-16-be', 'utf-16-le']

class EncryptionEncoding(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def rotate(self: EncryptionEncodingT, rotate_by: int) -> EncryptionEncodingT: ...
    def rotate_bruteforce(self: EncryptionEncodingT) -> EncryptionEncodingT: ...
    def rot_13(self: EncryptionEncodingT, amount: int=13, rotate_lower: bool=True, rotate_upper: bool=True, rotate_numbers: bool=False) -> EncryptionEncodingT: ...
    def rot_47(self: EncryptionEncodingT, amount: int=47) -> EncryptionEncodingT: ...
    def rot_47_bruteforce(self: EncryptionEncodingT) -> EncryptionEncodingT: ...
    def rot_8000(self: EncryptionEncodingT) -> EncryptionEncodingT: ...
    def xor(self: EncryptionEncodingT, key: Union[str, bytearray], key_type: Literal['hex', 'utf8', 'base64', 'decimal', 'raw']='hex') -> EncryptionEncodingT: ...
    def xor_bruteforce(self: EncryptionEncodingT, length: int=..., crib: Union[str, bytes, None]=...) -> EncryptionEncodingT: ...
    def jwt_decode(self: EncryptionEncodingT) -> EncryptionEncodingT: ...
    def jwt_verify(self: EncryptionEncodingT, secret: str, algorithm: list=...) -> EncryptionEncodingT: ...
    def jwt_sign(self: EncryptionEncodingT, secret: str, algorithms: str=...) -> EncryptionEncodingT: ...
    def jwt_token_generate_none_alg(self: EncryptionEncodingT, headers: Dict[str, Any]=...) -> EncryptionEncodingT: ...
    def jwt_token_generate_embedded_jwk(self: EncryptionEncodingT, private_key_pem: str, private_key_passphrase: str = ..., headers: dict = ..., alg: str = Literal["RS256", "RS512"]) -> EncryptionEncodingT: ...
    def rc4_encrypt(self: EncryptionEncodingT, key: str, key_format: RC4_FORMAT=...) -> EncryptionEncodingT: ...
    def rc4_decrypt(self: EncryptionEncodingT, key: str, key_format: RC4_FORMAT=...) -> EncryptionEncodingT: ...
    def des_encrypt(self: EncryptionEncodingT, key: str, iv: str=..., mode: Literal["CBC", "OFB", "CTR", "ECB"]="CBC", key_format: FORMAT="hex", iv_format: FORMAT="hex") -> EncryptionEncodingT: ...
    def des_decrypt(self: EncryptionEncodingT, key: str, iv: str=..., mode: Literal["CBC", "OFB", "CTR", "ECB"]="CBC", key_format: FORMAT="hex", iv_format: FORMAT="hex") -> EncryptionEncodingT: ...
    def salsa20_encrypt(self: EncryptionEncodingT, key: str, nonce: str=..., key_format: FORMAT="hex", nonce_format: FORMAT="hex") -> EncryptionEncodingT: ...
    def salsa20_decrypt(self: EncryptionEncodingT, key: str, nonce: str=..., key_format: FORMAT="hex", nonce_format: FORMAT="hex") -> EncryptionEncodingT: ...
    def chacha_encrypt(self: EncryptionEncodingT, key: str, nonce: str=..., key_format: FORMAT=..., nonce_format: FORMAT=...) -> EncryptionEncodingT: ...
    def chacha_decrypt(self: EncryptionEncodingT, key: str, nonce: str=..., key_format: FORMAT=..., nonce_format: FORMAT=...) -> EncryptionEncodingT: ...
    def triple_des_encrypt(self: EncryptionEncodingT, key: str, iv: str=..., mode: Literal["CBC", "OFB", "CTR", "ECB"]="CBC", key_format: FORMAT="hex", iv_format: FORMAT="hex") -> EncryptionEncodingT: ...
    def triple_des_decrypt(self: EncryptionEncodingT, key: str, iv: str=..., mode: Literal["CBC", "OFB", "CTR", "ECB", "ECB/NoPadding", "CBC/NoPadding"]="CBC", key_format: FORMAT="hex", iv_format: FORMAT="hex") -> EncryptionEncodingT: ...
    def aes_encrypt(self: EncryptionEncodingT, key: Union[bytes, str], iv: Union[bytes, str]=..., mode: Literal["CBC", "CFB", "OFB", "CTR", "ECB", "GCM"]="CBC", key_format: FORMAT="hex", iv_format: FORMAT="hex") -> EncryptionEncodingT: ...
    def aes_decrypt(self: EncryptionEncodingT, key: Union[bytes, str], iv: Union[bytes, str]=..., mode: Literal["CBC", "CFB", "OFB", "CTR", "ECB", "GCM", "ECB/NoPadding", "CBC/NoPadding"]="CBC", key_format: FORMAT="hex", iv_format: FORMAT="hex") -> EncryptionEncodingT: ...
    def blowfish_encrypt(self: EncryptionEncodingT, key: str, iv: str=..., mode: Literal["CBC", "OFB", "CTR", "ECB"]=..., key_format: FORMAT=..., iv_format: FORMAT=...) -> EncryptionEncodingT: ...
    def blowfish_decrypt(self: EncryptionEncodingT, key: str, iv: str=..., mode: Literal["CBC", "OFB", "CTR", "ECB"]=..., key_format: FORMAT=..., iv_format: FORMAT=...) -> EncryptionEncodingT: ...
    def vigenere_encode(self: EncryptionEncodingT, key: str) -> EncryptionEncodingT: ...
    def vigenere_decode(self: EncryptionEncodingT, key: str) -> EncryptionEncodingT: ...
    def affine_encode(self: EncryptionEncodingT, a: int=..., b: int=...) -> EncryptionEncodingT: ...
    def affine_decode(self: EncryptionEncodingT, a: int=..., b: int=...) -> EncryptionEncodingT: ...
    def atbash(self: EncryptionEncodingT) -> EncryptionEncodingT: ...
    def to_morse_code(self: EncryptionEncodingT, dot: str=..., dash: str=..., letter_delim: str=..., word_delim: str=...) -> EncryptionEncodingT: ...
    def from_morse_code(self: EncryptionEncodingT, dot: str=..., dash: str=..., letter_delim: str=..., word_delim: str=...) -> EncryptionEncodingT: ...
    def rsa_encrypt(self: EncryptionEncodingT, public_key: str, is_file: bool=False, passphrase: Union[str, None]=None, cipher:Literal['OAEP', 'PKCS']='OAEP') -> EncryptionEncodingT: ...
    def rsa_decrypt(self: EncryptionEncodingT, private_key: str, is_file: bool=False, passphrase: Union[str, None]=None, cipher:Literal['OAEP', 'PKCS']='OAEP') -> EncryptionEncodingT: ...
    def rsa_sign(self: EncryptionEncodingT, private_key: str, is_file: bool=False, passphrase: Union[str, None]=None, hash_format: Literal['SHA256', 'SHA512', 'SHA1', 'MD5', 'SHA384']='SHA256') -> EncryptionEncodingT: ...
    def rsa_verify(self: EncryptionEncodingT, signature: bytes, public_key: str, is_file: bool=False, passphrase: Union[str, None]=None, hash_format: Literal['SHA256', 'SHA512', 'SHA1', 'MD5', 'SHA384']='SHA256') -> EncryptionEncodingT: ...
    def rsa_private_pem_to_jwk(self: EncryptionEncodingT) -> EncryptionEncodingT: ...
    def rsa_public_key_from_jwk(self: EncryptionEncodingT) -> EncryptionEncodingT: ...
    def monoalphabetic_substitution(self: EncryptionEncodingT, mapping: Dict[str, str]=...): ...
    def to_letter_number_code(self: EncryptionEncodingT, join_by: Union[str, bytes]=...) -> EncryptionEncodingT: ...
    def from_letter_number_code(self: EncryptionEncodingT, delimiter: Union[str, bytes, None]=None, join_by: Union[str, bytes]=...) -> EncryptionEncodingT: ...
    def ls47_encrypt(self: EncryptionEncodingT, password: str, padding: int=..., signature: str=...) -> EncryptionEncodingT: ...
    def ls47_decrypt(self: EncryptionEncodingT, password: str, padding: int=...) -> EncryptionEncodingT: ...
    def bifid_encode(self: EncryptionEncodingT, key: Union[bytes, str]='') -> EncryptionEncodingT: ...
    def bifid_decode(self: EncryptionEncodingT, key: Union[bytes, str]='') -> EncryptionEncodingT: ...
    def huffman_encode(self: EncryptionEncodingT) -> EncryptionEncodingT: ...
    def huffman_decode(self: EncryptionEncodingT, huffman_codes: Dict[str, str]) -> EncryptionEncodingT: ...
    def cetacean_encode(self: EncryptionEncodingT) -> EncryptionEncodingT: ...
    def cetacean_decode(self: EncryptionEncodingT) -> EncryptionEncodingT: ...
    def rabbit(self: EncryptionEncodingT, key: str, iv: Union[None, str]=...) -> EncryptionEncodingT: ...
    def fernet_encrypt(self: EncryptionEncodingT, key:Union[bytes, str], encode_key: bool=False) -> EncryptionEncodingT: ...
    def fernet_decrypt(self: EncryptionEncodingT, key:Union[bytes, str], encode_key: bool=False) -> EncryptionEncodingT: ...
    def railfence_encode(self: EncryptionEncodingT, key: Union[int, str]=2, offset: Union[int, str]=0) -> EncryptionEncodingT: ...
    def railfence_decode(self: EncryptionEncodingT, key: Union[int, str]=2, offset: Union[int, str]=0) -> EncryptionEncodingT: ...
    def gpp_decrypt(self: EncryptionEncodingT) -> EncryptionEncodingT: ...
    def pgp_encrypt(self: EncryptionEncodingT, passphrase: Union[str, bytes], armoured: bool=False) -> EncryptionEncodingT: ...
    def pgp_decrypt(self: EncryptionEncodingT, passphrase: Union[str, bytes], armoured: bool=False) -> EncryptionEncodingT: ...
    def to_zeckendorf(self: EncryptionEncodingT, space_delim: Union[str, bytes] = ".", byte_delim: Union[str, bytes] = "|") -> EncryptionEncodingT: ...
    def from_zeckendorf(self: EncryptionEncodingT, space_delim: Union[str, bytes] = ".", byte_delim: Union[str, bytes] = "|") -> EncryptionEncodingT: ...
    def sms_encode_multitap(self: EncryptionEncodingT) -> EncryptionEncodingT: ...
    def sms_decode_multitap(self: EncryptionEncodingT) -> EncryptionEncodingT: ...


from ..core import ChepyCore
from typing import Any, TypeVar

OtherT = TypeVar('OtherT', bound='Other')

class Other(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def generate_uuid(self: OtherT) -> OtherT: ...



from typing import TypeVar, Union

import regex as re

from ..core import ChepyCore, ChepyDecorators

SearchT = TypeVar("SearchT", bound="Search")


class Search(ChepyCore):
    def __init__(self, *data):
        super().__init__(*data)

    """Class that is geared towards regex searches of secrets

    `Reference <https://github.com/dxa4481/truffleHog>`__
    """

    @ChepyDecorators.call_stack
    def search(self, pattern: Union[str, bytes]) -> SearchT:
        """Search. Group matches are returned as tuples.

        Args:
            pattern (Union[str, bytes]): Bytes pattern to search

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("abcdefg123 and again abcdefg123").search("abc(de)fg(12)(3)").o
            [('abcdefg123', 'de', '12', '3'), ('abcdefg123', 'de', '12', '3')]
        """
        pattern = self._str_to_bytes(pattern)
        self.state = re.findall(pattern, self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def search_list(self, pattern: Union[str, bytes]) -> SearchT:
        """Search all items in a list. List items are coerced into bytes first.
        Group matches are returned as tuples.

        Args:
            pattern (Union[str, bytes]): Bytes pattern to search

        Returns:
            Chepy: The Chepy object.
        """
        assert isinstance(self.state, list), "State is not a list"

        converted = [self._to_bytes(s) for s in self.state]
        pattern = self._str_to_bytes(pattern)
        pc = re.compile(pattern)

        hold = []
        for search in converted:
            matches = pc.findall(search)
            if len(matches) > 0:
                hold.append(matches)

        self.state = hold
        return self

    @ChepyDecorators.call_stack
    def search_ctf_flags(self, prefix: str, postfix: str = ".+?\\{*\\}") -> SearchT:
        """Search CTF style flags.

        This by default assumes that the flag format is similar
        to something like picoCTF{some_flag} as an example.

        Args:
            prefix (str): Prefix of the flag. Like `picoCTF`
            postfix (str, optional): Regex for the remainder of the flag.
                Defaults to '.+\\{.+\\}'.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("tests/files/flags").read_file().search_ctf_flags("pico").get_by_index(0)
            picoCTF{r3source_pag3_f1ag}
        """
        self.state = re.findall(prefix + postfix, self._convert_to_str(), re.IGNORECASE)
        return self

    @ChepyDecorators.call_stack
    def search_slack_tokens(self) -> SearchT:
        """Search slack tokens

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("tests/files/flags").read_file().search_slack_tokens().get_by_index(0)
            xoxp...859
        """
        self.state = re.findall(
            "(xox[p|b|o|a]-[0-9]{12}-[0-9]{12}-[0-9]{12}-[a-z0-9]{32})",
            self._convert_to_str(),
        )
        return self

    @ChepyDecorators.call_stack
    def search_slack_webhook(self) -> SearchT:
        """Search slack webhook

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(
            r"https://hooks\.slack\.com/services/T[a-zA-Z0-9_]{8}/B[a-zA-Z0-9_]{8}/[a-zA-Z0-9_]{24}",
            self._convert_to_str(),
        )
        return self

    @ChepyDecorators.call_stack
    def search_private_key(self) -> SearchT:
        """Search varios private key headers

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(
            "-----BEGIN (RSA|OPENSSH|DSA|EC) PRIVATE KEY-----", self._convert_to_str()
        )
        return self

    @ChepyDecorators.call_stack
    def search_twilio_key(self) -> SearchT:
        """Search for Twilio api key

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall("SK[a-z0-9]{32}", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def search_aws_key(self) -> SearchT:
        """Search for AWS key id

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall("AKIA[0-9A-Z]{16}", self._convert_to_str())
        return self



from ..core import ChepyCore
from typing import Any, TypeVar, Union
from typing_extensions import Literal as Literal

type_password_hashes = Literal[
    "apr_md5_crypt",
    "argon2",
    "atlassian_pbkdf2_sha1",
    "bcrypt",
    "bcrypt_sha256",
    "bigcrypt",
    "bsd_nthash",
    "bsdi_crypt",
    "cisco_asa",
    "cisco_pix",
    "cisco_type7",
    "crypt16",
    "cta_pbkdf2_sha1",
    "des_crypt",
    "django_digest",
    "dlitz_pbkdf2_sha1",
    "fshp",
    "grub_pbkdf2_sha512",
    "hex_digest",
    "ldap_crypt",
    "ldap_hex_md5",
    "ldap_hex_sha1",
    "ldap_md5",
    "ldap_pbkdf2_digest",
    "ldap_plaintext",
    "ldap_salted_md5",
    "ldap_salted_sha1",
    "ldap_salted_sha256",
    "ldap_salted_sha512",
    "ldap_sha1",
    "lmhash",
    "md5_crypt",
    "msdcc",
    "msdcc2",
    "mssql2000",
    "mssql2005",
    "mysql323",
    "mysql41",
    "nthash",
    "oracle10",
    "oracle11",
    "pbkdf2_digest",
    "phpass",
    "postgres_md5",
    "roundup_plaintext",
    "scram",
    "scrypt",
    "sha1_crypt",
    "sha256_crypt",
    "sha512_crypt",
    "sun_md5_crypt",
    "unix_disabled",
]

HashingT = TypeVar('HashingT', bound='Hashing')
hmac: Any
MD2: Any
MD4: Any
MD5: Any
SHA512: Any
SHA1: Any
SHA256: Any
keccak: Any
SHAKE128: Any
SHAKE256: Any
RIPEMD: Any
BLAKE2s: Any
BLAKE2b: Any
KDF: Any

class Hashing(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def sha1(self: HashingT) -> HashingT: ...
    def sha2_256(self: HashingT) -> HashingT: ...
    def sha2_512(self: HashingT) -> HashingT: ...
    def sha2_512_truncate(self: HashingT, truncate: Literal[256, 224]=...) -> HashingT: ...
    def sha2_384(self: HashingT) -> HashingT: ...
    def sha2_224(self: HashingT) -> HashingT: ...
    def sha3_512(self: HashingT) -> HashingT: ...
    def sha3_256(self: HashingT) -> HashingT: ...
    def sha3_384(self: HashingT) -> HashingT: ...
    def sha3_224(self: HashingT) -> HashingT: ...
    def md2(self: HashingT) -> HashingT: ...
    def md4(self: HashingT) -> HashingT: ...
    def md5(self: HashingT) -> HashingT: ...
    def keccak_512(self: HashingT) -> HashingT: ...
    def keccak_384(self: HashingT) -> HashingT: ...
    def keccak_256(self: HashingT) -> HashingT: ...
    def keccak_224(self: HashingT) -> HashingT: ...
    def shake_256(self: HashingT, size: int=...) -> HashingT: ...
    def shake_128(self: HashingT, size: int=...) -> HashingT: ...
    def ripemd_160(self: HashingT) -> HashingT: ...
    def blake_2b(self: HashingT, bits: Literal[512, 384, 256, 160, 128]=..., key: bytes=...) -> HashingT: ...
    def blake_2s(self: HashingT, bits: Literal[256, 160, 128]=..., key: bytes=...) -> HashingT: ...
    def crc8_checksum(self: HashingT) -> HashingT: ...
    def crc16_checksum(self: HashingT) -> HashingT: ...
    def crc32_checksum(self: HashingT) -> HashingT: ...
    def hmac_hash(self: HashingT, key: bytes=..., digest: Literal['md5', 'sha1', 'sha256', 'sha512']=...) -> HashingT: ...
    def derive_pbkdf2_key(self: HashingT, password: Union[str, bytes], salt: Union[str, bytes], key_size: int=..., iterations: int=..., hash_type: Literal['md5', 'sha1', 'sha256', 'sha512']=..., hex_salt: bool=..., show_full_key: bool=...) -> Any: ...
    def bcrypt_hash(self: HashingT, rounds: int=...) -> HashingT: ...
    def bcrypt_compare(self: HashingT, hash: str) -> HashingT: ...
    def scrypt_hash(self: HashingT, salt: str=..., key_length: int=..., N: int=..., r: int=..., p: int=...) -> HashingT: ...
    def password_hashing(self: HashingT, format: type_password_hashes, **kwargs) -> HashingT: ...



def yellow_background(s: str) -> str: ...
def red(s: str) -> str: ...
def blue(s: str) -> str: ...
def cyan(s: str) -> str: ...
def green(s: str) -> str: ...
def yellow(s: str) -> str: ...
def magenta(s: str) -> str: ...



from typing import List, Union
import binascii
import regex as re


class Base45:
    # reference: https://github.com/kirei/python-base45/blob/main/base45/__init__.py
    def __init__(self) -> None:
        self.BASE45_CHARSET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ $%*+-./:"
        self.BASE45_DICT = {v: i for i, v in enumerate(self.BASE45_CHARSET)}

    def b45encode(self, buf: bytes) -> bytes:
        """Convert bytes to base45-encoded string"""
        res = ""
        buflen = len(buf)
        for i in range(0, buflen & ~1, 2):
            x = (buf[i] << 8) + buf[i + 1]
            e, x = divmod(x, 45 * 45)
            d, c = divmod(x, 45)
            res += (
                self.BASE45_CHARSET[c] + self.BASE45_CHARSET[d] + self.BASE45_CHARSET[e]
            )
        if buflen & 1:
            d, c = divmod(buf[-1], 45)
            res += self.BASE45_CHARSET[c] + self.BASE45_CHARSET[d]
        return res.encode()

    def b45decode(self, s: Union[bytes, str]) -> bytes:
        """Decode base45-encoded string to bytes"""
        try:
            if isinstance(s, str):  # pragma: no cover
                buf = [self.BASE45_DICT[c] for c in s.rstrip("\n")]
            elif isinstance(s, bytes):
                buf = [self.BASE45_DICT[c] for c in s.decode()]
            else:  # pragma: no cover
                raise TypeError("Type must be 'str' or 'bytes'")

            buflen = len(buf)
            if buflen % 3 == 1:  # pragma: no cover
                raise ValueError("Invalid base45 string")

            res = []
            for i in range(0, buflen, 3):
                if buflen - i >= 3:
                    x = buf[i] + buf[i + 1] * 45 + buf[i + 2] * 45 * 45
                    if x > 0xFFFF:  # pragma: no cover
                        raise ValueError
                    res.extend(divmod(x, 256))
                else:
                    x = buf[i] + buf[i + 1] * 45
                    if x > 0xFF:  # pragma: no cover
                        raise ValueError
                    res.append(x)
            return bytes(res)
        except (ValueError, KeyError, AttributeError):  # pragma: no cover
            raise ValueError("Invalid base45 string")


class Base92(object):
    """
    Reference: https://github.com/Gu-f/py3base92/tree/master
    """

    CHARACTER_SET = r"!#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_abcdefghijklmnopqrstuvwxyz{|}"

    @classmethod
    def base92_chr(cls, val):
        if val < 0 or val >= 91:  # pragma: no cover
            raise ValueError("val must be in [0, 91)")
        if val == 0:
            return "!"  # pragma: no cover
        elif val <= 61:
            return chr(ord("#") + val - 1)
        else:
            return chr(ord("a") + val - 62)

    @classmethod
    def base92_ord(cls, val):
        num = ord(val)
        if val == "!":
            return 0  # pragma: no cover
        elif ord("#") <= num and num <= ord("_"):
            return num - ord("#") + 1
        elif ord("a") <= num and num <= ord("}"):
            return num - ord("a") + 62
        else:  # pragma: no cover
            raise ValueError("val is not a base92 character")

    @classmethod
    def b92encode(cls, byt: bytes) -> str:
        if not isinstance(byt, bytes):  # pragma: no cover
            raise TypeError(f"a bytes-like object is required, not '{type(byt)}'")
        if not byt:
            return "~"
        if not isinstance(byt, str):
            byt = "".join([chr(b) for b in byt])
        bitstr = ""
        while len(bitstr) < 13 and byt:
            bitstr += "{:08b}".format(ord(byt[0]))
            byt = byt[1:]
        resstr = ""
        while len(bitstr) > 13 or byt:
            i = int(bitstr[:13], 2)
            resstr += cls.base92_chr(i // 91)
            resstr += cls.base92_chr(i % 91)
            bitstr = bitstr[13:]
            while len(bitstr) < 13 and byt:
                bitstr += "{:08b}".format(ord(byt[0]))
                byt = byt[1:]

        if bitstr:
            if len(bitstr) < 7:
                bitstr += "0" * (6 - len(bitstr))
                resstr += cls.base92_chr(int(bitstr, 2))
            else:  # pragma: no cover
                bitstr += "0" * (13 - len(bitstr))
                i = int(bitstr, 2)
                resstr += cls.base92_chr(i // 91)
                resstr += cls.base92_chr(i % 91)
        return resstr

    @classmethod
    def b92decode(cls, bstr: str) -> bytes:
        if not isinstance(bstr, str):  # pragma: no cover
            raise TypeError(f"a str object is required, not '{type(bstr)}'")
        bitstr = ""
        resstr = ""
        if bstr == "~":
            return "".encode(encoding="latin-1")

        for i in range(len(bstr) // 2):
            x = cls.base92_ord(bstr[2 * i]) * 91 + cls.base92_ord(bstr[2 * i + 1])
            bitstr += "{:013b}".format(x)
            while 8 <= len(bitstr):
                resstr += chr(int(bitstr[0:8], 2))
                bitstr = bitstr[8:]
        if len(bstr) % 2 == 1:
            x = cls.base92_ord(bstr[-1])
            bitstr += "{:06b}".format(x)
            while 8 <= len(bitstr):
                resstr += chr(int(bitstr[0:8], 2))
                bitstr = bitstr[8:]
        return resstr.encode(encoding="latin-1")


class LZ77Compressor:
    """
    Class containing compress and decompress methods using LZ77 compression algorithm.
    Reference: https://the-algorithms.com/algorithm/lz-77?lang=python
    """

    def __init__(self, window_size: int = 13, lookahead_buffer_size: int = 6) -> None:
        self.window_size = window_size
        self.lookahead_buffer_size = lookahead_buffer_size
        self.search_buffer_size = self.window_size - self.lookahead_buffer_size

    def compress(self, text: str) -> list:
        output = []
        search_buffer = ""

        while text:
            token = self._find_encoding_token(text, search_buffer)

            search_buffer += text[: token[1] + 1]
            if len(search_buffer) > self.search_buffer_size:
                search_buffer = search_buffer[-self.search_buffer_size :]

            text = text[token[1] + 1 :]

            output.append(token)

        return output

    def decompress(self, tokens: list) -> str:
        output = ""

        for token in tokens:
            for _ in range(token[1]):
                output += output[-token[0]]
            output += token[2]

        return output

    def _find_encoding_token(self, text: str, search_buffer: str):
        if not text:
            raise ValueError("We need some text to work with.")  # pragma: no cover

        length, offset = 0, 0

        if not search_buffer:
            return [offset, length, text[length]]

        for i, character in enumerate(search_buffer):  # pragma: no cover
            found_offset = len(search_buffer) - i
            if character == text[0]:
                found_length = self._match_length_from_index(text, search_buffer, 0, i)
                if found_length >= length:
                    offset, length = found_offset, found_length

        return [offset, length, text[length]]  # pragma: no cover

    def _match_length_from_index(
        self, text: str, window: str, text_index: int, window_index: int
    ) -> int:
        if not text or text[text_index] != window[window_index]:  # pragma: no cover
            return 0
        return 1 + self._match_length_from_index(
            text, window + text[text_index], text_index + 1, window_index + 1
        )  # pragma: no cover


class UUEncoderDecoder:
    def __init__(self, data: bytes, header: str = "-"):
        self.data = data
        self.header = header

    def split_data(self, data, chunk_size=45):
        for i in range(0, len(data), chunk_size):
            yield self.data[i : i + chunk_size]

    def uuencode(self):
        encoded_chunks = []
        for chunk in self.split_data(self.data):
            encoded_data = binascii.b2a_uu(chunk)
            encoded_chunks.append(encoded_data.decode("utf-8"))

        # UUencode header and footer
        header = f"begin 644 {self.header}\n"
        footer = " \nend\n"

        return header + "\n".join(encoded_chunks) + footer

    def uudecode(self):
        lines = self.data.strip().split(b"\n")

        # More flexible header check - just look for "begin"
        if len(lines) < 3 or not lines[0].lower().startswith(b"begin"):
            raise ValueError(
                "Invalid UUencode format. Missing header"
            )  # pragma: no cover

        # Find where data actually starts (skip header)
        data_start = 1

        # Find where data ends (before 'end' or backtick line)
        data_end = len(lines)
        for i in range(len(lines) - 1, 0, -1):
            if lines[i].strip().lower() == b"end":
                data_end = i
                break
            elif (
                lines[i].strip() == b"`"
            ):  # Handle backtick before end  # pragma: no cover
                data_end = i
                break

        data_lines = lines[data_start:data_end]

        decoded_data = []
        for line in data_lines:
            # Skip empty lines and backticks
            if not line.strip() or line.strip() == b"`":
                continue
            try:
                decoded_chunk = binascii.a2b_uu(line)
                decoded_data.append(decoded_chunk)
            except binascii.Error:  # pragma: no cover
                # Skip lines that can't be decoded (like extra backticks)
                continue

        return b"".join(decoded_data)


class Uint1Array:
    # reference https://git.sr.ht/~evanhahn/UTF-21.js
    def __init__(self, bit_length_or_buffer):
        if isinstance(bit_length_or_buffer, int):
            bit_length = bit_length_or_buffer
            self.bit_length = bit_length
            self.bytes = bytearray((bit_length + 7) // 8)
        else:
            buffer = bit_length_or_buffer
            self.bit_length = len(buffer) * 8
            self.bytes = bytearray(buffer)

    @property
    def buffer(self):
        return bytes(self.bytes)

    def get(self, index):
        byte_index = index // 8
        bit_index = index % 8
        byte = self.bytes[byte_index]
        return (byte >> (7 - bit_index)) & 1

    def set(self, index, value):
        byte_index = index // 8
        bit_index = index % 8
        old_byte = self.bytes[byte_index]
        new_byte = old_byte | (value << (7 - bit_index))
        self.bytes[byte_index] = new_byte


def detect_delimiter(
    data: Union[str, bytes],
    delimiters: List[Union[str, bytes]] = [
        "; ",
        ";",
        ". ",
        ".",
        "- ",
        "-",
        "\\",
        ": ",
        ":",
        "/",
        ", ",
        ",",
        " ",
        "\n",
        "\\x",
        "\\0x",
    ],
    default_delimiter: str = "",
) -> Union[str, bytes, None]:
    """Detect delimiter

    Args:
        data (Union[str, bytes]): Data
        delimiters (List[Union[str, bytes]], optional): Array of delimiters. Defaults to [" ", ";", ".", "-", "\"].
        default_delimiter (str): The default delimiter

    Returns:
        Union[str, bytes, None]: Delimiter or None if one is not found
    """
    if default_delimiter:
        return default_delimiter

    is_bytes = False
    if isinstance(data, bytes):  # pragma: no cover
        delimiters = [d.encode() for d in delimiters]
        is_bytes = True

    for delimiter in delimiters:
        parts = data.split(delimiter)
        if len(parts) > 1 and all(part.strip() for part in parts):
            return delimiter

    if default_delimiter:  # pragma: no cover
        return default_delimiter.encode() if is_bytes else default_delimiter
    else:
        return None


class Rotate:
    def __init__(self, data: bytes, radix: int):
        self.data = data
        self.radix = radix

    def rot(self, algo):
        result = []
        for byte in self.data:
            b = byte
            for _ in range(self.radix):
                b = algo(b)
            result.append(b)
        return b"".join([chr(x).encode() for x in result])

    def rot_right_carry(self):
        result = []
        carryBits = 0

        amount = self.radix % 8
        for i in range(len(self.data)):
            oldByte = self.data[i] & 0xFF  # Ensure it's treated as an unsigned byte
            newByte = (oldByte >> amount) | carryBits
            carryBits = (oldByte & ((1 << amount) - 1)) << (8 - amount)
            result.append(newByte)

        result[0] |= carryBits
        return b"".join([chr(x).encode() for x in result])

    @staticmethod
    def rotate_right(b):
        bit = (b & 1) << 7
        return (b >> 1) | bit

    @staticmethod
    def rotate_left(b):
        bit = (b >> 7) & 1
        return ((b << 1) | bit) & 0xFF

    def rotate_left_carry(self):
        result = bytearray(len(self.data))
        carryBits = 0

        amount = self.radix % 8
        for i in range(len(self.data) - 1, -1, -1):
            oldByte = self.data[i]
            newByte = ((oldByte << amount) | carryBits) & 0xFF
            carryBits = (oldByte >> (8 - amount)) & ((1 << amount) - 1)
            result[i] = newByte

        result[-1] |= carryBits

        return b"".join([chr(x).encode() for x in result])


class _Base64:
    base_64_chars = {
        "standard": "A-Za-z0-9+/=",
        "url_safe": "A-Za-z0-9-_",
        "filename_safe": "A-Za-z0-9+\\-=",
        "itoa64": "./0-9A-Za-z=",
        "xml": "A-Za-z0-9_.",
        # "y64": "A-Za-z0-9._-",
        "z64": "0-9a-zA-Z+/=",
        "radix_64": "0-9A-Za-z+/=",
        # "uuencoding": " -_",
        "xxencoding": "+\\-0-9A-Za-z",
        # "binHex": "!-,-0-689@A-NP-VX-Z[`a-fh-mp-r",
        "rot13": "N-ZA-Mn-za-m0-9+/=",
        "unix_crypt": "./0-9A-Za-z",
        # "atom128": "/128GhIoPQROSTeUbADfgHijKLM+n0pFWXY456xyzB7=39VaqrstJklmNuZvwcdEC",
        # "megan35": "3GHIJKLMNOPQRSTUb=cdefghijklmnopWXYZ/12+406789VaqrstuvwxyzABCDEF5",
        # "zong22": "ZKj9n+yf0wDVX1s/5YbdxSo=ILaUpPBCHg8uvNO4klm6iJGhQ7eFrWczAMEq3RTt2",
        # "hazz15": "HNO4klm6ij9n+J2hyf0gzA8uvwDEq3X1Q7ZKeFrWcVTts/MRGYbdxSo=ILaUpPBC5",
    }

    @staticmethod
    def decode_base64(data, alphabet):
        output = []
        i = 0

        # Calculate the necessary padding
        padding_required = (4 - len(data) % 4) % 4
        data += padding_required * "="

        while i < len(data):
            enc1 = alphabet.index(data[i]) if i < len(data) and data[i] != "=" else 0
            i += 1
            enc2 = alphabet.index(data[i]) if i < len(data) and data[i] != "=" else 0
            i += 1
            enc3 = alphabet.index(data[i]) if i < len(data) and data[i] != "=" else 0
            i += 1
            enc4 = alphabet.index(data[i]) if i < len(data) and data[i] != "=" else 0
            i += 1

            chr1 = (enc1 << 2) | (enc2 >> 4)
            chr2 = ((enc2 & 15) << 4) | (enc3 >> 2)
            chr3 = ((enc3 & 3) << 6) | enc4

            if 0 <= chr1 < 256:
                output.append(chr1)
            if 0 <= chr2 < 256 and data[i - 2] != "=":
                output.append(chr2)
            if 0 <= chr3 < 256 and data[i - 1] != "=":
                output.append(chr3)

        return bytes(output)

    @staticmethod
    def encode_base64(data: bytes, alphabet: str):
        output = ""
        i = 0
        padding_char = (
            "=" if alphabet[-1] == "=" else None
        )  # Check if '=' is in the alphabet, otherwise use None

        while i < len(data):
            chr1 = data[i] if i < len(data) else 0
            i += 1
            chr2 = data[i] if i < len(data) else 0
            i += 1
            chr3 = data[i] if i < len(data) else 0
            i += 1

            enc1 = chr1 >> 2
            enc2 = ((chr1 & 3) << 4) | (chr2 >> 4)
            enc3 = ((chr2 & 15) << 2) | (chr3 >> 6)
            enc4 = chr3 & 63

            if i > len(data) + 1:
                enc3 = 64
                enc4 = 64
            elif i > len(data):
                enc4 = 64

            output += alphabet[enc1]
            output += alphabet[enc2]
            output += (
                alphabet[enc3]
                if enc3 < 64
                else (padding_char if padding_char is not None else "")
            )
            output += (
                alphabet[enc4]
                if enc4 < 64
                else (padding_char if padding_char is not None else "")
            )

        # Remove padding characters if they are not part of the alphabet
        if padding_char is None:
            output = output.rstrip(
                alphabet[-1]
            )  # Strip the last character of the alphabet if it's not '='

        return output


def expand_alpha_range(alph_str: str, join_by: Union[str, None] = None):
    def expand_range(start, end):
        return [str(x) for x in range(int(start), int(end) + 1)]

    def expand_char_range(start, end):
        return [chr(x) for x in range(ord(start), ord(end) + 1)]

    hold = []
    i = 0
    length = len(alph_str)

    while i < length:
        # Check for numeric ranges
        if (
            i < length - 2
            and alph_str[i].isdigit()
            and alph_str[i + 1] == "-"
            and alph_str[i + 2].isdigit()
        ):
            start = ""
            while i < length and alph_str[i].isdigit():
                start += alph_str[i]
                i += 1
            i += 1  # Skip the '-'
            end = ""
            while i < length and alph_str[i].isdigit():
                end += alph_str[i]
                i += 1
            hold.extend(expand_range(start, end))
        elif (
            i < length - 2
            and alph_str[i].isalpha()
            and alph_str[i + 1] == "-"
            and alph_str[i + 2].isalpha()
        ):
            start = alph_str[i]
            end = alph_str[i + 2]
            hold.extend(expand_char_range(start, end))
            i += 3
        elif (
            i < length - 2 and alph_str[i] == "\\" and alph_str[i + 1] == "-"
        ):  # pragma: no cover
            hold.append("-")
            i += 2
        else:
            hold.append(alph_str[i])
            i += 1

    if join_by is not None:
        return join_by.join(hold)
    return hold


class Zeckendorf:
    @staticmethod
    def generate_fibonacci_numbers(max_val):
        """Generate Fibonacci numbers up to max_val, starting from F(2)=1, F(3)=2, etc."""
        fibs = [1, 2]  # Start with F(2)=1, F(3)=2
        while fibs[-1] < max_val:
            next_fib = fibs[-1] + fibs[-2]
            if next_fib > max_val:
                break
            fibs.append(next_fib)
        return fibs

    @staticmethod
    def to_zeckendorf(n):
        """Convert a positive integer to its Zeckendorf representation (non-neighbouring Fibonacci sum)."""
        if n <= 0:
            return []  # pragma: no cover

        # Generate Fibonacci numbers up to n
        fibs = Zeckendorf.generate_fibonacci_numbers(n)

        result = []
        i = len(fibs) - 1

        # Greedy algorithm: use largest possible Fibonacci numbers
        while n > 0 and i >= 0:
            if fibs[i] <= n:
                result.append(fibs[i])
                n -= fibs[i]
                # Skip the immediate previous Fibonacci number to ensure non-neighbouring
                i -= 2
            else:
                i -= 1

        return sorted(result, reverse=True)



from ..core import ChepyCore
from typing import Any, TypeVar

DateTimeT = TypeVar("DateTimeT", bound="DateTime")

class DateTime(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def from_unix_timestamp(self: DateTimeT, format: str = ..., utc: bool = ...) -> DateTimeT: ...
    def to_unix_timestamp(self: DateTimeT) -> DateTimeT: ...



import json
import lazy_import
from typing import Iterator, Dict, List, Union
from binascii import hexlify, unhexlify
from itertools import cycle
from urllib.request import urlopen
RSA = lazy_import.lazy_module("Crypto.PublicKey.RSA")

from .combinatons import generate_combo, hex_chars
from chepy import Chepy


def factordb(n: int) -> dict:  # pragma: no cover
    """Query the factordb api and get primes if available

    Args:
        n (int): n is the modulus for the public key and the private keys

    Returns:
        dict: response from api as a dictionary. None if status code is not 200
    """
    res = urlopen("http://factordb.com/api/?query={}".format(str(n)))
    if res.status != 200:
        return None
    return json.loads(res.read().decode())


def construct_private_key(
    n: int, e: int, d: int, format: str = "PEM", passphrase: str = None
) -> str:  # pragma: no cover
    """Construct a private key given n, e and d

    Args:
        n (int): n
        e (int): e
        d (int): d
        format (str, optional): Supports PEM, DER and OpenSSH. Defaults to "PEM".
        passphrase (str, optional): [description]. Defaults to None.

    Returns:
        str: Private key
    """
    valid_formats = ["PEM", "DER", "OpenSSH"]
    assert format in valid_formats, "Valid formats are {}".format(
        " ".join(valid_formats)
    )
    priv = RSA.construct((n, e, d))
    return priv.export_key(format=format, passphrase=passphrase)


def xor_bruteforce_multi(
    data: str, min: int = 0, max: int = None, errors: str = "backslashreplace"
) -> Iterator[Dict[str, str]]:
    """Bruteforce multibyte xor encryption. For faster results, use pypy3.
    It is important to set the min and max values if the key size is known.

    Args:
        data (str): XORed data
        min (int, optional): Minimum key length. Default will start at 1 byte
            . Defaults to 0.
        max (int, optional): Maximum key length. Maximum value is 257 bytes. Defaults to None.
        errors (str, optional): How should the errors be handled? Defaults to backslashreplace.
            Valid options are replace, ignore, backslashreplace

    Returns:
        Iterator[Dict[str, str]]: A dictionary where key is key, and value is xored data

    Yields:
        Iterator[Dict[str, str]]: A generator which contains a dictionary with the
            keys: `key` and `out`
    """
    for key in generate_combo(
        hex_chars(), min_length=min, max_length=max, join_by=""
    ):  # pragma: no cover
        yield {
            "key": key,
            "out": Chepy(data).xor(key).bytearray_to_str(errors=errors).o,
        }


def xor_repeating_key(
    data1: bytes, data2: bytes, min: int = 1, max: int = 257
) -> Union[bytes, None]:  # pragma: no cover
    """Recover repeating key xor keys.

    Args:
        data1 (bytes): File 1 path
        data2 (bytes): File 2 path
        min (int, optional): Min chars to test. Defaults to 1.
        max (int, optional): Max chars to test. Defaults to 257.

    Returns:
        Union[bytes, None]: Key as hex bytes or None if no key found
    """

    def find_same(s: bytes):
        i = (s + s).find(s, 1, -1)
        return None if i == -1 else s[:i]

    for i in range(min, max):
        d1 = data1[:i]

        d2 = data2[:i]

        x = bytes(a ^ b for a, b in zip(d1, d2))
        o = find_same(x)
        if o is not None:
            return o


def xor(data: bytes, key: bytes) -> bytes:  # pragma: no cover
    """XOR data with a hex key

    Args:
        data (bytes): Data to be xored
        key (bytes): Hex key to xor data with

    Returns:
        bytes: XORed data

    Example:
        >>> xor(b"hello", unhexlify(b"aabbccdd"))
        b'c2dea0b1c5'
    """
    return hexlify(bytes(a ^ b for a, b in zip(data, cycle(key))))


def one_time_pad_crib(
    cipherText1: Union[bytes, str], cipherText2: Union[bytes, str], crib: bytes
) -> List[str]:
    """One time pad crib attack.

    Args:
        cipherText1 (Union[bytes, str]): Cipher text 1 as hex
        cipherText2 (Union[bytes, str]): Cipher text 2 as hex
        crib (bytes): Crib (known text) as bytes

    Returns:
        List[str]: List of possible plaintexts
    """
    cipherText1 = unhexlify(cipherText1)
    cipherText2 = unhexlify(cipherText2)
    xored = bytearray(a ^ b for a, b in zip(cipherText1, cipherText2))
    hold = []
    for offset in range(0, len(xored) - len(crib) + 1):
        piece = xored[offset : offset + len(crib)]
        piece = bytearray(a ^ b for a, b in zip(crib, piece))
        if all(32 <= c <= 126 for c in piece):
            piece = (
                ("." * offset)
                + piece.decode()
                + ("." * (len(xored) - len(crib) - offset))
            )
            hold.append(piece)
    return hold


def generate_rsa_keypair(
    bits: int = 1024, passphrase: str = None
) -> Dict[str, dict]:  # pragma: no cover
    """Generates an RSA keypair with the specified number of bits.

    Args:
      bits: The number of bits for the RSA keypair.

    Returns:
      A tuple of the RSA public key and RSA private key, both in PEM format.
    """

    keypair = RSA.generate(bits)
    return {
        "pem": {
            "public": keypair.publickey().exportKey("PEM"),
            "private": keypair.exportKey("PEM", passphrase=passphrase),
        },
        "der": {
            "public": keypair.publickey().exportKey("DER"),
            "private": keypair.exportKey("DER", passphrase=passphrase),
        },
    }



import lazy_import
import binascii
import base64
import codecs
import html
import base58
import json
import struct
import pickle
import string
import itertools
import quopri
import io
import csv
from six import indexbytes, int2byte, unichr
import sqlite3
import collections
from random import randint
import regex as re
from .internal.constants import Encoding, Base65536
from .internal.helpers import (
    detect_delimiter,
    Rotate,
    Uint1Array,
    UUEncoderDecoder,
    Base92,
    Base45,
    _Base64,
    expand_alpha_range,
)

yaml = lazy_import.lazy_module("yaml")
import regex as re
import hexdump
from ast import literal_eval
from typing import TypeVar, Union, List, Literal
from urllib.parse import quote_plus as _urllib_quote_plus
from urllib.parse import unquote_plus as _urllib_unquote_plus

crypto_number = lazy_import.lazy_module("Crypto.Util.number")
msgpack = lazy_import.lazy_module("msgpack")

from ..core import ChepyCore, ChepyDecorators
from chepy.modules.internal.constants import Encoding
import chepy.modules.internal.rison as rison

DataFormatT = TypeVar("DataFormatT", bound="DataFormat")


class DataFormat(ChepyCore):
    def __init__(self, *data):
        super().__init__(*data)

    @ChepyDecorators.call_stack
    def eval_state(self) -> DataFormatT:
        """Eval state as python.
        Handy when converting string representation
        of objects.

        Returns:
            Chepy: The Chepy object
        """
        self.state = literal_eval(self.state)
        return self

    @ChepyDecorators.call_stack
    def list_to_bytes(self, ascii: bool = False) -> DataFormatT:
        """Convert bytes (array of bytes) to ascii

        Args:
            ascii (bool, optional): Output as ascii, by default False

        Returns:
            Chepy: The Chepy object.
        """
        assert isinstance(self.state, list), "Data in state is not a list"
        if ascii:
            self.state = bytearray(self.state).decode()
        else:
            self.state = bytes(bytearray(self.state))
        return self

    @ChepyDecorators.call_stack
    def list_to_str(self, join_by: Union[str, bytes] = " ") -> DataFormatT:
        """Join an array by `join_by`

        Args:
            join_by (Union[str, bytes], optional): String character to join by, by default ' '

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy(["a", "b", "c"]).list_to_str(",").o
            "a,b,c"
        """
        assert isinstance(self.state, list), "Data in state not a list"
        # convert the list of items in state appropriately
        if isinstance(join_by, str):
            self.state = [str(x) for x in self.state]
        elif isinstance(join_by, bytes):
            self.state = [bytes(x) for x in self.state]
        self.state = join_by.join(self.state)
        return self

    @ChepyDecorators.call_stack
    def str_list_to_list(self) -> DataFormatT:
        """Convert a string list to a list

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("[1,2,'lol', true]").str_list_to_list().o
            [1, 2, "lol", True]
        """
        self.state = json.loads(re.sub(r"'", '"', self._convert_to_str()))
        return self

    @ChepyDecorators.call_stack
    def join(self, join_by: Union[str, bytes] = "") -> DataFormatT:
        """Join a list with specified character

        Args:
            join_by (Union[str, bytes], optional): What to join with. Defaults to ""

        Returns:
            Chepy: The Chepy object.
        Examples:
            >>> Chepy(["a", "b", "c"]).join_list(":").o
            "a:b:c"
        """
        assert isinstance(self.state, list), "State is not a list"
        data = [self._to_bytes(x) for x in self.state]
        join_by = self._str_to_bytes(join_by)
        self.state = join_by.join(data)
        return self

    @ChepyDecorators.call_stack
    def json_to_dict(self) -> DataFormatT:
        """Convert a JSON string to a dict object

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy('{"some": "data", "a": ["list", 1, true]}').json_to_dict().o
            {
                "some": "data",
                "a": ["list", 1, True],
            }
        """
        self.state = json.loads(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def dict_to_json(self) -> DataFormatT:
        """Convert a dict object to a JSON string

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy({"some": "data", "a": ["list", 1, True]}).dict_to_json().o
            '{"some":"data","a":["list",1,true]}'
        """
        assert isinstance(self.state, dict), "Not a dict object"
        self.state = json.dumps(self.state)
        return self

    @ChepyDecorators.call_stack
    def dict_get_items(self, *keys: str) -> DataFormatT:
        """Get items from a dict. If no keys are specified, it will return all items.
        Returns:
            Chepy: The Chepy object.
        Examples:
            >>> o = Chepy({"a": 1, "b": 2}).dict_get_items("a", "b", "c").o
            [1, 2]
        """
        assert isinstance(self.state, dict), "Not a dict object"
        if len(keys) == 0:
            self.state = list(self.state.values())
            return self
        o = list()
        for k in keys:
            if self.state.get(k):
                o.append(self.state.get(k))
        self.state = o
        return self

    @ChepyDecorators.call_stack
    def yaml_to_json(self) -> DataFormatT:  # pragma: no cover
        """Convert yaml to a json string

        Returns:
            Chepy: The Chepy object.
        """
        self.state = json.dumps(yaml.safe_load(self.state))
        return self

    @ChepyDecorators.call_stack
    def json_to_yaml(self) -> DataFormatT:
        """Convert a json string to yaml structure

        Returns:
            Chepy: The Chepy object.
        """

        class ChepyYamlDumper(yaml.Dumper):
            def increase_indent(self, flow=False, indentless=False):
                return super(ChepyYamlDumper, self).increase_indent(flow, False)

        self.state = yaml.dump(
            json.loads(self.state),
            Dumper=ChepyYamlDumper,
            default_flow_style=False,
            sort_keys=False,
            allow_unicode=True,
        )
        return self

    @ChepyDecorators.call_stack
    def to_base58(self) -> DataFormatT:
        """Encode as Base58

        Base58 is a notation for encoding arbitrary byte data using a
        restricted set of symbols that can be conveniently used by humans
        and processed by computers.This property encodes raw data
        into an ASCII Base58 string.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some data").to_base58().out.decode()
            "2UDrs31qcWSPi"
        """
        self.state = base58.b58encode(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def from_base58(self) -> DataFormatT:
        """Decode as Base58

        Base58 is a notation for encoding arbitrary byte data using a
        restricted set of symbols that can be conveniently used by humans
        and processed by computers.This property decodes raw data
        into an ASCII Base58 string.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("2UDrs31qcWSPi").from_base58().out.decode()
            "some data"
        """
        self.state = base58.b58decode(self.state)
        return self

    @ChepyDecorators.call_stack
    def to_base85(self) -> DataFormatT:
        """Encode as Base58

        Base85 is a notation for encoding arbitrary byte data using a
        restricted set of symbols that can be conveniently used by humans
        and processed by computers.This property decodes raw data
        into an ASCII Base58 string.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some data").to_base85().out.decode()
            "F)Po,+Cno&@/"
        """
        self.state = base64.a85encode(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def from_base85(self) -> DataFormatT:
        """Decode as Base85

        Base85 is a notation for encoding arbitrary byte data using a
        restricted set of symbols that can be conveniently used by humans
        and processed by computers.This property decodes raw data
        into an ASCII Base58 string.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("F)Po,+Cno&@/").from_base85().out.decode()
            "some data"
        """
        self.state = base64.a85decode(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def to_base16(self) -> DataFormatT:
        """Encode state in base16

        Returns:
            Chepy: The Chepy object.
        """
        self.state = base64.b16encode(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def from_base16(self) -> DataFormatT:
        """Decode state in base16

        Returns:
            Chepy: The Chepy object.
        """
        self.state = base64.b16decode(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def to_base32(self) -> DataFormatT:
        """Encode as Base32

        Base32 is a notation for encoding arbitrary byte data using a
        restricted set of symbols that can be conveniently used by humans
        and processed by computers. It uses a smaller set of characters than
        Base64, usually the uppercase alphabet and the numbers 2 to 7.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some data").base32_encode().out.decode()
            "ONXW2ZJAMRQXIYI="
        """
        self.state = base64.b32encode(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def from_base32(self, remove_whitespace: bool = True) -> DataFormatT:
        """Decode as Base32

        Base32 is a notation for encoding arbitrary byte data using a
        restricted set of symbols that can be conveniently used by humans
        and processed by computers. It uses a smaller set of characters than
        Base64, usually the uppercase alphabet and the numbers 2 to 7.

        Args:
            remove_whitespace(bool, optional): If true, all whitespaces are removed

        Returns:
            Chepy: The Chepy object.
        """
        if remove_whitespace:
            self.state = self.remove_whitespace().o
        self.state = base64.b32decode(self.state)
        return self

    @ChepyDecorators.call_stack
    def to_base92(self) -> DataFormatT:
        """Encode to Base92

        Returns:
            Chepy: The Chepy object.
        """
        self.state = Base92.b92encode(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def from_base92(self) -> DataFormatT:
        """Decode from Base92

        Returns:
            Chepy: The Chepy object.
        """
        self.state = Base92.b92decode(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def to_base45(self) -> DataFormatT:
        """Encode to Base45

        Returns:
            Chepy: The Chepy object.
        """
        self.state = Base45().b45encode(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def from_base45(self) -> DataFormatT:
        """Decode from Base45

        Returns:
            Chepy: The Chepy object.
        """
        self.state = Base45().b45decode(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def to_base91(self) -> DataFormatT:  # pragma: no cover
        """Base91 encode
        Reference: https://github.com/aberaud/base91-python/blob/master/base91.py#L69

        Returns:
            Chepy: The Chepy object.
        """
        bindata = self._convert_to_bytes()
        b = 0
        n = 0
        out = ""
        for count in range(len(bindata)):
            byte = bindata[count : count + 1]
            b |= struct.unpack("B", byte)[0] << n
            n += 8
            if n > 13:
                v = b & 8191
                if v > 88:
                    b >>= 13
                    n -= 13
                else:
                    v = b & 16383
                    b >>= 14
                    n -= 14
                out += (
                    Encoding.BASE91_ALPHABET[v % 91] + Encoding.BASE91_ALPHABET[v // 91]
                )
        if n:
            out += Encoding.BASE91_ALPHABET[b % 91]
            if n > 7 or b > 90:
                out += Encoding.BASE91_ALPHABET[b // 91]
        self.state = out
        return self

    @ChepyDecorators.call_stack
    def from_base91(self) -> DataFormatT:  # pragma: no cover
        """Decode as Base91
        Reference: https://github.com/aberaud/base91-python/blob/master/base91.py#L42

        Returns:
            Chepy: The Chepy object.
        """
        encoded_str = self._convert_to_str()
        decode_table = dict((v, k) for k, v in enumerate(Encoding.BASE91_ALPHABET))
        v = -1
        b = 0
        n = 0
        out = bytearray()
        for strletter in encoded_str:
            if strletter not in decode_table:
                continue
            c = decode_table[strletter]
            if v < 0:
                v = c
            else:
                v += c * 91
                b |= v << n
                n += 13 if (v & 8191) > 88 else 14
                while True:
                    out += struct.pack("B", b & 255)
                    b >>= 8
                    n -= 8
                    if not n > 7:
                        break
                v = -1
        if v + 1:
            out += struct.pack("B", (b | v << n) & 255)
        self.state = out
        return self

    @ChepyDecorators.call_stack
    def to_int(
        self, byteorder: Literal["little", "big"] = "big", base: int = 10
    ) -> DataFormatT:
        """Converts the string representation of a number into an int

        Args:
            byteorder(Literal['little', 'big']): Byte order if state is bytes
            base(int): If state is a str, the base to convert it to.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("1").to_int().o
            1
        """
        if isinstance(self.state, bytes):
            self.state = int.from_bytes(self.state, byteorder)
        elif isinstance(self.state, str):
            self.state = int(self.state, int(base))
        return self

    @ChepyDecorators.call_stack
    def to_bytes(self) -> DataFormatT:
        """Converts the data in state to bytes

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy({'some': 'val', 'kl': 1}).to_bytes().o
            b"{'some': 'val', 'kl': 1}"
        """
        self.state = self._convert_to_str().encode()
        return self

    @ChepyDecorators.call_stack
    def from_bytes(self) -> DataFormatT:
        """Decodes bytes to string.

        Returns:
            Chepy: The Chepy object.
        """
        self.state = self._convert_to_bytes().decode()
        return self

    @ChepyDecorators.call_stack
    def to_base64(self, alphabet: str = "standard") -> DataFormatT:
        """Encode as Base64

        Base64 is a notation for encoding arbitrary byte data using a
        restricted set of symbols that can be conveniently used by humans
        and processed by computers.This property encodes raw data
        into an ASCII Base64 string.

        Args:
            alphabet (str, optional): Provide a custom charset to base64 with. Valid values are: filename_safe, itoa64, radix_64, rot13, standard, unix_crypt, url_safe, xml, xxencoding, z64

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> # To use a custom character set, use:
            >>> custom = "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
            >>> Chepy("Some data").to_base64(custom=custom).o
            b'IqxhNG/YMLFV'
        """
        data = self._convert_to_bytes()
        alphabet = alphabet.strip()

        char_set = expand_alpha_range(
            _Base64.base_64_chars.get(alphabet, alphabet), join_by=""
        )
        if len(char_set) < 63 or len(char_set) > 66:  # pragma: no cover
            raise ValueError(
                "Invalid base64 chars. Should be 63-66 chars. " + str(len(char_set))
            )

        self.state = _Base64.encode_base64(data, alphabet=char_set)
        return self

    @ChepyDecorators.call_stack
    def from_base64(
        self, alphabet: str = "standard", remove_non_alpha: bool = True
    ) -> DataFormatT:
        """Decode as Base64

        Base64 is a notation for encoding arbitrary byte data using a
        restricted set of symbols that can be conveniently used by humans
        and processed by computers.This property decodes raw data
        into an ASCII Base64 string.

        Args:
            alphabet (str, optional): Provide a custom charset to base64 with. Valid values are: filename_safe, itoa64, radix_64, rot13, standard, unix_crypt, url_safe, xml, xxencoding, z64
            remove_whitespace(bool, optional): If true, all whitespaces are removed (Defaults to True)
            remove_non_alpha(bool, optional): If true, all whitespaces are removed. (Defaults to True)

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("QqxhNG/mMKtYPqoz64FVR42=")
            >>> c.from_base64(alphabet="./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz")
            >>> c.out
            b"some random? data"
        """
        alphabet = alphabet.strip()
        char_set = expand_alpha_range(
            _Base64.base_64_chars.get(alphabet, alphabet), join_by=""
        )
        if len(char_set) < 63 or len(char_set) > 65:  # pragma: no cover
            raise ValueError(
                "Invalid base64 chars. Should be 63-65 chars. " + str(len(char_set))
            )

        data = self._convert_to_str()

        if remove_non_alpha:
            data = re.sub("[^" + char_set + "]", "", data)

        # if is_standard or alphabet == 'url_safe':
        #     data += "=="
        padding_needed = len(data) % 4
        if padding_needed and alphabet != "url_safe":
            data += "=" * (4 - padding_needed)

        # if is_standard:
        #     self.state = base64.b64decode(data)
        # if alphabet == 'url_safe':
        #     self.state = base64.urlsafe_b64decode(data)
        # else:
        self.state = _Base64.decode_base64(data, char_set)
        return self

    @ChepyDecorators.call_stack
    def decode_bytes(self, errors: str = "ignore") -> DataFormatT:
        """Decode bytes to string

        Args:
            errors (str, optional): Ignore or replace error chars. Defaults to 'ignore'.

        Returns:
            Chepy: The Chepy object.
        """
        self.state = self._convert_to_bytes().decode(errors=errors)
        return self

    @ChepyDecorators.call_stack
    def to_hex(self, delimiter: str = "") -> DataFormatT:
        """Converts a string to its hex representation

        Args:
            delimiter (str, optional): Delimiter. Defaults to None.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("AAA").to_hex().out.decode()
            "414141"
        """
        if delimiter == "":
            self.state = binascii.hexlify(self._convert_to_bytes())
        else:
            self.state = binascii.hexlify(self._convert_to_bytes(), sep=delimiter)
        return self

    @ChepyDecorators.call_stack
    def from_hex(
        self,
        delimiter: str = None,
        join_by: str = "",
        replace: Union[bytes, None] = b"%|0x",
    ) -> DataFormatT:
        """Convert a non delimited hex string to string

        Args:
            delimiter (str, optional): Delimiter. Defaults to None.
            join_by (str, optional): Join by. Defaults to ' '.
            replace (Union[bytes, None], optional): Regex pattern to replace hex string prefixes. Defaults to b'%x|0x'.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("414141").from_hex().out
            b"AAA"
        """
        data = self._convert_to_bytes()
        if replace is not None:
            replace = self._str_to_bytes(replace)
            data = re.sub(replace, b"", data)
        if delimiter is None:
            delimiter = detect_delimiter(data, default_delimiter=None)
        if delimiter is not None:
            self.state = join_by.encode().join(
                list(
                    binascii.unhexlify(x)
                    for x in data.split(self._str_to_bytes(delimiter))
                )
            )
        else:
            self.state = binascii.unhexlify(data)
        return self

    @ChepyDecorators.call_stack
    def hex_to_int(self) -> DataFormatT:
        """Converts hex into its integer representation

        Returns:
            Chepy: The Chepy object.

        Examples:
            Chepy works with hex characters that start with a 0x

            >>> Chepy("0x123").hex_to_int().out
            291

            Without 0x in the hex

            >>> Chepy("123").hex_to_int().out
            291
        """
        if self._convert_to_str().startswith("0x"):
            self.state = int(self.state, 0)
        else:
            self.state = int(self.state, 16)
        return self

    @ChepyDecorators.call_stack
    def hex_to_bytes(self) -> DataFormatT:
        """Hex to bytes hex

        Converts a hex string to its bytes form. Example:
        41 becomes \\x41

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("ab00").hex_to_bytes().o
            b"\\xab\\x00"
        """
        self.state = binascii.unhexlify(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def hex_to_str(self, ignore: bool = False) -> DataFormatT:
        """Decodes a hex string to ascii ignoring any decoding errors

        Args:
            ignore (bool, optional): Ignore errors, by default False

        Returns:
            Chepy: The Chepy object.

        Examples:
            To ignore UnicodeDecode errors, set ignore to True
            >>> Chepy("4100").hex_to_str(ignore=True).o
            "A\x00"
        """
        if ignore:
            self.state = binascii.unhexlify(self._convert_to_bytes()).decode(
                errors="ignore"
            )
        else:
            self.state = binascii.unhexlify(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def str_to_hex(self, delimiter: Union[str, bytes] = b"") -> DataFormatT:
        """Converts a string to a hex string

        Args:
            delimiter (Union[str, bytes], optional): Format delimiter. Defaults to b'

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_bytes()
        delimiter = self._bytes_to_str(delimiter)
        out = delimiter + delimiter.join(f"{byte:02x}" for byte in data)
        self.state = out
        return self

    @ChepyDecorators.call_stack
    def int_to_bytes(self, length: Union[int, None] = None, byteorder: str = "big"):
        """Int to bytes conversion

        Args:
            length (int, optional): Length of bytes. If not specified, byteorder is ignored and long_to_bytes is used.
            byteorder (str, optional): Endianness. Defaults to 'big'.

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_int()
        if not length:
            self.state = crypto_number.long_to_bytes(self.state)
        else:
            self.state = data.to_bytes(int(length), byteorder=byteorder)
        return self

    @ChepyDecorators.call_stack
    def int_to_hex(self, unsigned_int: bool = True, bit_size: int = 64) -> DataFormatT:
        """Converts an integer into its hex equivalent

        Args:
            unsigned_int (bool, optional): Calculate the unsigned hexadecimal representation. Defaults to True
            bit_size (int, optional): Used when unsinged_int is used

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy(101).int_to_hex().o
            "65"
        """
        i = self._convert_to_int()
        if unsigned_int and i < 0:
            self.state = hex((1 << bit_size) + i).strip("0x").strip("-0x")
        else:
            self.state = format(self._convert_to_int(), "x")
        return self

    @ChepyDecorators.call_stack
    def int_to_str(self) -> DataFormatT:
        """Converts an integer into a string

        Returns:
            Chepy: The Chepy object.
        """
        self.state = self._convert_to_str()
        return self

    @ChepyDecorators.call_stack
    def binary_to_hex(self) -> DataFormatT:
        """Converts binary data into a hex string

        Returns:
            Chepy: The Chepy object.
        """
        self.state = binascii.hexlify(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def normalize_hex(self, is_bytearray=False) -> DataFormatT:
        """Normalize a hex string

        Removes special encoding characters from a hex string like %,
        0x, , :, ;, \\n and \\r\\n

        Args:
            is_bytearray (bool, optional): Set to True if state is a bytearray

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("41:42:CE").normalize_hex().o
            "4142CE"
            >>> Chepy("0x410x420xce").normalize_hex().o
            "4142ce"
        """
        if is_bytearray:
            self.state = binascii.hexlify(bytearray(self.state))
            return self
        else:
            delimiters = [" ", "0x", "%", ",", ";", ":", r"\\n", "\\r\\n"]
            string = re.sub("|".join(delimiters), "", self.state)
            self.state = string
            return self

    @ChepyDecorators.call_stack
    def str_from_hexdump(self) -> DataFormatT:
        """Extract a string from a hexdump

        Returns:
            Chepy: The Chepy object.
        """
        # TODO make new line aware \n \r\n \0a etc
        self.state = "".join(re.findall(r"\|(.+)\|", self._convert_to_str()))
        return self

    @ChepyDecorators.call_stack
    def to_hexdump(self) -> DataFormatT:
        """Convert the state to hexdump

        Returns:
            Chepy: The Chepy object.
        """
        self.state = hexdump.hexdump(self._convert_to_bytes(), result="return")
        return self

    @ChepyDecorators.call_stack
    def from_hexdump(self) -> DataFormatT:
        """Convert hexdump back to str

        Returns:
            Chepy: The Chepy object.
        """
        self.state = hexdump.restore(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def to_url_encoding(self, safe: str = "", all_chars: bool = False) -> DataFormatT:
        """URL encode

        Encodes problematic characters into percent-encoding,
        a format supported by URIs/URLs.

        Args:
            safe (str, optional): String of characters that will not be encoded, by default ""
            all_chars (bool, optional): Encode all characters including safe characters

        Returns:
            Chepy: The Chepy object.

        Examples:
            Url encode while specifying save characters

            >>> Chepy("https://google.com/?lol=some data&a=1").to_url_encoding(safe="/:").o
            "https://google.com/%3Flol%3Dsome+data%26a%3D1"
        """
        data = self._convert_to_str()

        def encode_all(string):
            return "".join("%{0:0>2x}".format(ord(char)) for char in string)

        if all_chars:
            self.state = encode_all(data)
        else:
            self.state = _urllib_quote_plus(data, safe=safe)
        return self

    @ChepyDecorators.call_stack
    def from_url_encoding(self) -> DataFormatT:
        """Converts URI/URL percent-encoded characters back to their raw values.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("https://google.com/%3Flol%3Dsome+data%26a%3D1").from_url_encoding().o
            "https://google.com/?lol=some data&a=1"
        """
        self.state = _urllib_unquote_plus(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def bytearray_to_str(
        self, encoding: str = "utf8", errors: str = "replace"
    ) -> DataFormatT:
        """Convert a python bytearray to string

        Args:
            encoding (str, optional): String encoding. Defaults to 'utf8'.
            errors (str, optional): How errors should be handled. Defaults to replace.

        Raises:
            TypeError: If state is not a bytearray

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy(bytearray("lolol", "utf")).bytearray_to_str().o
            "lolol"
        """
        if isinstance(self.state, bytearray):
            self.state = self.state.decode(encoding, errors=errors)
            return self
        else:  # pragma: no cover
            raise TypeError("State is not a bytearray")

    @ChepyDecorators.call_stack
    def to_list(self) -> DataFormatT:
        """Convert string to list

        Converts the string in state to an array of individual characyers

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("abc").to_list().o
            ["a", "b", "c"]
        """
        if isinstance(self.state, (bytearray, bytes, str)):
            self.state = list(self.state)
        else:  # pragma: no cover
            raise ValueError("State is not a valid data type")
        return self

    @ChepyDecorators.call_stack
    def str_to_dict(self) -> DataFormatT:
        """Convert string to a dictionary

        Returns:
            Chepy: The Chepy object.
        """
        self.state = yaml.safe_load(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def to_charcode(self, join_by: str = " ", base: int = 16) -> DataFormatT:
        """Convert a string to a list of unicode charcode

        Converts text to its unicode character code equivalent.
        e.g. Γειά σου becomes 0393 03b5 03b9 03ac 20 03c3 03bf 03c5

        Args:
            join_by (str, optional): String to join the charcodes by. Defaults to ' '.
            base (int, optional): Base to use for the charcodes. Defaults to 16.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("aㅎ").to_charcode()
            "61 314e"
        """
        hold = []
        for c in self._convert_to_str():
            hold.append(str(int(hex(ord(c))[2:], base)))
        self.state = join_by.join(hold)
        return self

    @ChepyDecorators.call_stack
    def from_charcode(
        self, delimiter: str = None, join_by: str = "", base: int = 10
    ) -> DataFormatT:
        """Convert array of unicode chars to string

        Args:
            delimiter (str, optional): Delimiter. Defaults to " ".
            join_by (str, optional): Join by. Defaults to "".
            base (int, optional): Base. Defaults to 10.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("314e 61 20 41"]).from_charcode().o
            "ㅎa A"
        """
        data = self._convert_to_str()
        out = []
        if not delimiter:
            delimiter = detect_delimiter(data)
        for c in data.split(delimiter):
            out.append(chr(int(c, base)))
        self.state = join_by.join(out)
        return self

    @ChepyDecorators.call_stack
    def to_decimal(self, join_by: str = " ") -> DataFormatT:
        """Convert characters to decimal

        Args:
            join_by (str, optional): Join the decimal values by this. Defaults to ' '.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("aㅎ").to_decimal().o
            '97 12622'
        """
        self.state = join_by.join(
            str(x) for x in list(ord(s) for s in list(self._convert_to_str()))
        )
        return self

    @ChepyDecorators.call_stack
    def from_decimal(self, delimiter: str = None, join_by: str = "") -> DataFormatT:
        """Convert a list of decimal numbers to string

        Args:
            delimiter (str, optional): Delimiter. Defaults to " ".
            join_by (str, optional): Join by. Defaults to "".

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy(12622).from_decimal().o
            "ㅎ"
        """
        data = self._convert_to_str()
        if not delimiter:
            delimiter = detect_delimiter(data)
        self.state = join_by.join(
            list(chr(int(s)) for s in data.strip().split(delimiter))
        )
        return self

    @ChepyDecorators.call_stack
    def to_binary(
        self, join_by: Union[str, bytes] = " ", byte_length: int = 8
    ) -> DataFormatT:
        """Convert string characters to binary

        Args:
            join_by (str, optional): join_by. Defaults to " ".
            byte_length (int, optional): Byte length. Defaults to 8.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("abc").to_binary().o
            b"01100001 01100010 01100011"
        """
        hold = []
        join_by = self._str_to_bytes(join_by)
        # out = list(format(s, "08b").encode() for s in list(self._convert_to_bytes()))
        for s in list(self._convert_to_bytes()):
            hold.append(str(bin(s)[2:].zfill(byte_length)).encode())
        self.state = join_by.join(hold)
        return self

    @ChepyDecorators.call_stack
    def from_binary(self, delimiter: str = None, byte_length: int = 8) -> DataFormatT:
        """Convert a list of binary numbers to string

        Args:
            delimiter (str, optional): Delimiter. Defaults to " ".
            byte_length (int, optional): Byte length. Defaults to 8.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("01100001 01100010 01100011").from_binary().o
            "abc"
        """
        data = self._convert_to_str()
        if not delimiter:
            delimiter = detect_delimiter(data)
        n = int(
            "".join([x[byte_length - 8 :] for x in data.split(delimiter)]),
            2,
        )
        self.state = n.to_bytes((n.bit_length() + 7) // 8, "big")
        return self

    @ChepyDecorators.call_stack
    def to_octal(self, join_by: str = " ") -> DataFormatT:
        """Convert string characters to octal

        Args:
            join_by (str, optional): Join by. Defaults to "".

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("abㅎ").to_octal().o
            "141 142 30516"
        """
        self.state = join_by.join(
            list(format(ord(s), "0o") for s in list(self._convert_to_str()))
        )
        return self

    @ChepyDecorators.call_stack
    def from_octal(self, delimiter: str = None, join_by: str = "") -> DataFormatT:
        """Convert a list of octal numbers to string

        Args:
            delimiter (str, optional): Delimiter. Defaults to None.
            join_by (str, optional): Join by. Defaults to "".

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("141 142").from_octal().o
            "ab"
        """
        data = self._convert_to_str()
        if not delimiter:
            delimiter = detect_delimiter(data, default_delimiter=delimiter)
        self.state = join_by.join(
            list(chr(int(str(x), 8)) for x in data.split(delimiter))
        )
        return self

    @ChepyDecorators.call_stack
    def to_html_entity(self, format="named", all_chars=False) -> DataFormatT:
        """Encode html entities

        Encode special html characters like & > < etc

        Args:
            format (str): Encoding format. Valid formats are named, numeric and hex. Defaults to named
            all_chars (bool): If all chars should be encoded. By default a-ZA-Z0-9 are skipped. Defaults to False

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy('https://google.com&a="lol"').to_html_entity().o
            "https://google.com&amp;a=&quot;lol&quot;"
        """
        data = self._convert_to_str()
        chars = {k: 1 for k in string.ascii_letters + string.digits}
        hold = ""
        for d in data:
            if not all_chars and chars.get(d) is not None:
                hold += d
                continue
            if format == "named":
                hold += Encoding.BYTE_TO_ENTITY.get(ord(d), f"&#{ord(d)};")
            elif format == "hex":
                hold += f"&#x{d.encode().hex()};"
            elif format == "numeric":
                hold += f"&#{ord(d)};"
        self.state = hold
        return self

    @ChepyDecorators.call_stack
    def from_html_entity(self) -> DataFormatT:
        """Decode html entities

        Decode special html characters like & > < etc

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("https://google.com&amp;a=&quot;lol&quot;").from_html_entity().o
            'https://google.com&a="lol"'
        """
        self.state = html.unescape(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def to_punycode(self) -> DataFormatT:
        """Encode to punycode

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("münchen").to_punycode().o
            b"mnchen-3ya"
        """
        self.state = self._convert_to_str().encode("punycode")
        return self

    @ChepyDecorators.call_stack
    def from_punycode(self) -> DataFormatT:
        """Decode to punycode

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy(b"mnchen-3ya").from_punycode().o
            "münchen"
        """
        self.state = self._convert_to_bytes().decode("punycode")
        return self

    @ChepyDecorators.call_stack
    def encode_bruteforce(self) -> DataFormatT:
        """Bruteforce the various encoding for a string

        Enumerates all supported text encodings for the input,
        allowing you to quickly spot the correct one.
        `Reference <https://docs.python.org/2.4/lib/standard-encodings.html>`__

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("münchen한").encode_bruteforce()
            {
                'ascii': b'm\\xfcnchen\\ud55c',
                'base64_codec': b'bcO8bmNoZW7tlZw=\\n',
                'big5': b'm\\xfcnchen\\ud55c',
                'big5hkscs': b'm\\x88\\xa2nchen\\ud55c',
                ...
            }
        """
        data = self._convert_to_str()
        final = dict()
        for enc in Encoding.py_encodings:
            final[enc] = data.encode(enc, errors="backslashreplace")

        for text_enc in Encoding.py_text_encodings:
            try:
                final[text_enc] = codecs.encode(data, text_enc)
            except TypeError:
                final[text_enc] = codecs.encode(data.encode(), text_enc)
            except UnicodeEncodeError:
                try:
                    final[text_enc] = codecs.encode(
                        data, text_enc, errors="backslashreplace"
                    )
                except TypeError:  # pragma: no cover
                    final[text_enc] = codecs.encode(
                        data.encode(), text_enc, errors="backslashreplace"
                    )
        self.state = final
        return self

    @ChepyDecorators.call_stack
    def decode_bruteforce(self) -> DataFormatT:
        """Bruteforce the various decoding for a string

        Enumerates all supported text encodings for the input,
        allowing you to quickly spot the correct one.
        `Reference <https://docs.python.org/2.4/lib/standard-encodings.html>`__

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("m\\xfcnchen\\ud55c").decode_bruteforce()
            {
                ...
                'unicode_escape': 'münchen한',
                'utf_16': '屭晸湣档湥畜㕤挵',
                'utf_16_be': '浜硦据捨敮屵搵㕣',
                ...
            }
        """
        data = self._convert_to_bytes()
        final = dict()
        for enc in Encoding.py_encodings:
            final[enc] = data.decode(enc, errors="backslashreplace")

        for text_enc in Encoding.py_text_encodings:
            try:
                final[text_enc] = codecs.decode(
                    data, text_enc, errors="backslashreplace"
                )
            except UnicodeDecodeError:  # pragma: no cover
                final[text_enc] = codecs.decode(
                    data.decode(), text_enc, errors="backslashreplace"
                )
            except AssertionError:
                final[text_enc] = ""
                continue
            except UnicodeError:
                final[text_enc] = ""
                continue
            except TypeError:
                final[text_enc] = ""
                continue
        self.state = final
        return self

    @ChepyDecorators.call_stack
    def to_braille(self) -> DataFormatT:
        """Convert text to six-dot braille symbols

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("secret message").to_braille().o
            "⠎⠑⠉⠗⠑⠞⠀⠍⠑⠎⠎⠁⠛⠑"
        """
        chars = dict(zip(Encoding.asciichars, Encoding.brailles))
        self.state = "".join(list(chars.get(c.lower()) for c in self.state))
        return self

    @ChepyDecorators.call_stack
    def from_braille(self) -> DataFormatT:
        """Convert text to six-dot braille symbols

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("⠎⠑⠉⠗⠑⠞⠀⠍⠑⠎⠎⠁⠛⠑").from_braille().o
            "secret message"
        """
        chars = dict(zip(Encoding.brailles, Encoding.asciichars))
        self.state = "".join(list(chars.get(c.lower()) for c in self.state))
        return self

    @ChepyDecorators.call_stack
    def trim(self) -> DataFormatT:
        """Trim string. Removes whitespaces

        Returns:
            Chepy: The Chepy object.
        """
        self.state = self._convert_to_str().strip()
        return self

    @ChepyDecorators.call_stack
    def to_nato(self, join_by: str = " ") -> DataFormatT:
        """Convert string to NATO phonetic format.

        Example: abc = Alpha Bravo Charlie

        Args:
            join_by (str, optional): [description]. Defaults to " ".

        Returns:
            Chepy: The Chepy object
        """
        nato_chars = Encoding.NATO_CONSTANTS_DICT
        hold = []
        data: str = self._convert_to_str()
        for d in data:
            if d.isalpha():
                hold.append(nato_chars.get(d.upper(), d.upper()))
            else:
                hold.append(d)
        self.state = join_by.join(hold)
        return self

    @ChepyDecorators.call_stack
    def from_nato(
        self, delimiter: Union[str, None] = None, join_by: str = ""
    ) -> DataFormatT:
        """Translate NATO phoentic to words

        Args:
            delimiter (str, optional): Delimiter to split on. Defaults to ' '.
            join_by (str, optional): Join result by. Defaults to ''.

        Returns:
            Chepy: The Chepy object
        """
        data = self._convert_to_str()
        if delimiter is None:
            delimiter = detect_delimiter(data)
        data = data.split(delimiter)
        d = {v: k for k, v in Encoding.NATO_CONSTANTS_DICT.items()}
        self.state = join_by.join([d.get(p, p) for p in data])
        return self

    @ChepyDecorators.call_stack
    def swap_values(
        self, indices1: Union[str, List[int]], indices2: Union[str, List[int]]
    ):
        """Swaps the values in the global bytes based on indices specified in two lists.

        Args:
            indices1 (Union[str, List[int]]): The first indices. If string, it should be , delimited
            indices2 (Union[str, List[int]]): The second indices. If string, it should be , delimited

        Returns:
            Chepy: The Chepy object.
        """
        if isinstance(indices1, str):
            indices1 = [int(x.strip()) for x in indices1.split(",")]
        if isinstance(indices2, str):
            indices2 = [int(x.strip()) for x in indices2.split(",")]

        is_list = isinstance(self.state, list)
        if is_list:
            data = self.state
        else:
            data = bytearray(self._convert_to_bytes())

        # Check if both lists are of the same length
        if len(indices1) != len(indices2):  # pragma: no cover
            raise ValueError("The two argument lists must have the same length.")

        # Swap values in the byte list
        for idx1, idx2 in zip(indices1, indices2):
            # Validate indices are within bounds
            if (
                idx1 < 0 or idx1 >= len(data) or idx2 < 0 or idx2 >= len(data)
            ):  # pragma: no cover
                raise IndexError(
                    f"Index out of range. Lenth of state is {len(data)}, indices provided {idx1} {idx2}"
                )

            # Perform the swap
            data[idx1], data[idx2] = data[idx2], data[idx1]

        # Convert the modified list back to bytes
        self.state = data if is_list else bytes(data)
        return self

    @ChepyDecorators.call_stack
    def swap_strings(self, by: int) -> DataFormatT:
        """Swap characters in string

        Args:
            by (int): Number of bytes to swap

        Returns:
            Chepy: The Chepy object
        """
        t = list(self.state)
        t[::by], t[1::by] = t[1::by], t[::by]
        self.state = "".join(t)
        return self

    @ChepyDecorators.call_stack
    def to_string(self) -> DataFormatT:
        """Convert to string

        Returns:
            Chepy: The Chepy object
        """
        self.state = self._convert_to_str()
        return self

    @ChepyDecorators.call_stack
    def stringify(self, compact: bool = True) -> DataFormatT:
        """Stringify the state. This uses json.dumps unlike to_string

        Args:
            compact (bool, optional): If the output should be compact. Defaults to True.

        Returns:
            Chepy: The Chepy object.
        """
        sep = None
        if compact:
            sep = (",", ":")
        self.state = json.dumps(self.state, separators=sep)
        return self

    @ChepyDecorators.call_stack
    def select(self, start: Union[int, str, bytes], end: int = None) -> DataFormatT:
        """Get an item by specifying an index

        Args:
            start (int): Starting index number to get
            end (int, optional): Ending index number to get. If none specified, will be end of item. Defaults to None.

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_bytes()
        if isinstance(
            start,
            (
                str,
                bytes,
            ),
        ):
            pattern = self._to_bytes(start)
            s = re.search(pattern, data)
            if s is not None:
                start = s.start()
            else:  # pragma: no cover
                raise ValueError("Pattern not found")
        if end is None:
            self.state = data[start:]
        else:
            self.state = data[start:end]
        return self

    @ChepyDecorators.call_stack
    def length(self) -> DataFormatT:
        """Get the length of the current state as string

        Returns:
            Chepy: The Chepy object.
        """
        self.state = len(self.state)
        return self

    @ChepyDecorators.call_stack
    def to_leetcode(self, replace_space: str = "") -> DataFormatT:
        """Convert to leetcode. Reference
        Reference github.com/ss8651twtw/CTF-flag-generator

        Args:
            replace_space (str, optional): Replace spaces with specified char. Defaults to ''.

        Returns:
            Chepy: The Chepy object.
        """

        def change(c):
            if replace_space and c == " ":
                return replace_space
            if c.isalpha():
                c = c.upper()
                char_set = Encoding.LEETCODE[ord(c) - ord("A")]
                new_c = char_set[randint(0, len(char_set) - 1)]
                return new_c
            else:
                return c

        hold = ""
        string = self._convert_to_str()
        for s in string:
            hold += change(s)
        self.state = hold
        return self

    @ChepyDecorators.call_stack
    def substitute(self, x: str, y: str) -> DataFormatT:
        """Replace a subset of specified characters in the state.

        Args:
            x (str): Chars to replace
            y (str): Chars to replace with

        Returns:
            Chepy: The Chepy object.
        """
        assert len(x) == len(y), "x and y chars are not of equal length"
        s = self._convert_to_str()
        o = s.maketrans(x, y)
        self.state = s.translate(o)
        return self

    @ChepyDecorators.call_stack
    def remove(self, pattern: Union[str, bytes] = b"") -> DataFormatT:
        """Remove is identifcal to find_replace, except it removes the identified
        pattern with an empty value.

        Args:
            pattern (Union[str, bytes], optional): Replace non-printable characters with this. Defaults to ''.

        Returns:
            Chepy: The Chepy object.
        """
        self.find_replace(pattern, b"")
        return self

    @ChepyDecorators.call_stack
    def remove_nonprintable(self, replace_with: Union[str, bytes] = b"") -> DataFormatT:
        """Remove non-printable characters from string.

        Args:
            replace_with (bytes, optional): Replace non-printable characters with this. Defaults to ''.

        Returns:
            Chepy: The Chepy object.
        """
        replace_with = self._str_to_bytes(replace_with)
        data = self._convert_to_bytes()
        self.state = re.sub(b"[^[:print:]]", replace_with, data)
        return self

    @ChepyDecorators.call_stack
    def swap_endianness(
        self, word_length: int = 4, pad_incomplete: bool = True
    ) -> DataFormatT:
        """Swap endianness.

        Args:
            word_length (int, optional): Word length. Use 8 for big endian. Defaults to 4.
            pad_incomplete (bool, optional): Pad incomplete word. Defaults to True.

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_bytes()
        num_bytes = len(data)
        padding_length = (word_length - num_bytes) % word_length
        padded_data = data + b"\x00" * padding_length

        swapped_data = b""
        for i in range(0, len(padded_data), word_length):
            word = padded_data[i : i + word_length]
            swapped_word = struct.unpack("<" + "B" * word_length, word)[::-1]
            swapped_data += struct.pack("B" * word_length, *swapped_word)

        if not pad_incomplete:
            swapped_data = swapped_data.replace(b"\x00", b"")
        self.state = swapped_data
        return self

    @ChepyDecorators.call_stack
    def bruteforce_from_base_xx(self) -> DataFormatT:
        """Bruteforce various base encodings. Current supports base85, base16, base32, base64, base85, base58

        Returns:
            Chepy: The Chepy object.
        """
        hold = {}
        ops = {
            "base85": base64.a85decode,
            "base16": base64.b16decode,
            "base32": base64.b32decode,
            "base64": base64.b64decode,
            "base58": base58.b58decode,
        }
        data = self._convert_to_bytes()
        for do in ops.items():
            try:
                hold[do[0]] = do[1](data)
            except:
                hold[do[0]] = None
        self.state = hold
        return self

    @ChepyDecorators.call_stack
    def long_to_bytes(self) -> DataFormatT:
        """Long numbers to bytes

        Returns:
            Chepy: The Chepy object.
        """
        d = self._convert_to_int()
        self.state = crypto_number.long_to_bytes(d)
        return self

    @ChepyDecorators.call_stack
    def bytes_to_long(self) -> DataFormatT:
        """Bytes to long

        Returns:
            Chepy: The Chepy object.
        """
        d = self._convert_to_bytes()
        self.state = crypto_number.bytes_to_long(d)
        return self

    @ChepyDecorators.call_stack
    def concat(self, data: Union[str, bytes]) -> DataFormatT:
        """Concat bytes to the current state

        Args:
            data (Union[str, bytes]): Data to add

        Returns:
            Chepy: The Chepy object. s
        """
        data = self._str_to_bytes(data)
        self.state = self._convert_to_bytes() + data
        return self

    @ChepyDecorators.call_stack
    def to_wingdings(self) -> DataFormatT:
        """Encode to windings

        Returns:
            Chepy: The Chepy object.
        """
        hold = ""
        for c in list(self._convert_to_str()):
            hold += chr(Encoding.wingdings.get(c, ord(c)))
        self.state = hold.encode()
        return self

    @ChepyDecorators.call_stack
    def from_wingdings(self) -> DataFormatT:
        """Decode from windings

        Returns:
            Chepy: The Chepy object.
        """
        conv = {v: k for k, v in Encoding.wingdings.items()}
        hold = ""
        for i in list(self._convert_to_str()):
            hold += conv.get(ord(i), i)
        self.state = hold.encode()
        return self

    @ChepyDecorators.call_stack
    def from_twin_hex(self) -> DataFormatT:
        """Decode twin hex

        Returns:
            Chepy: The Chepy object.
        """
        spaceChar = " "
        cypherBase = []
        for x in range(32, 128):
            for y in range(32, 128):
                thisPair = chr(x) + chr(y)
                cypherBase.append(thisPair)

        def tripleSplit(strInput):
            outArray = []
            thisTriple = ""
            i = 0
            while i < len(strInput):
                thisTriple = strInput[i]
                i += 1
                if i < len(strInput):
                    thisTriple += strInput[i]
                else:  # pragma: no cover
                    thisTriple += spaceChar
                i += 1
                if i < len(strInput):
                    thisTriple += strInput[i]
                else:  # pragma: no cover
                    thisTriple += spaceChar
                outArray.append(thisTriple)
                i += 1
            return outArray

        def base36_encode(string):
            return int(string, 36)

        inputArray = tripleSplit(self._convert_to_str())
        strOutput = ""
        thisPair = ""
        for code in inputArray:
            if code and len(code):
                thisCode = base36_encode(code)
                thisPair = cypherBase[thisCode]
                strOutput += thisPair
        self.state = strOutput.encode()
        return self

    @ChepyDecorators.call_stack
    def to_twin_hex(self) -> DataFormatT:
        """Encode to twin hex encoding

        Returns:
            Chepy: The Chepy object.
        """

        def base36_decode(number):
            alphabet = "0123456789abcdefghijklmnopqrstuvwxyz"
            if number == 0:  # pragma: no cover
                return "0"
            base36 = ""
            while number != 0:
                number, i = divmod(number, 36)
                base36 = alphabet[i] + base36
            return base36

        def dHex(thisPair):
            intCode = cypherBase.index(thisPair)
            strOutput = base36_decode(intCode)
            if len(strOutput) == 2:  # pragma: no cover
                return strOutput + spaceChar
            else:
                return strOutput

        spaceChar = " "
        cypherBase = []
        for x in range(32, 128):
            for y in range(32, 128):
                thisPair = chr(x) + chr(y)
                cypherBase.append(thisPair)

        strInput = self._convert_to_str()
        strOutput = ""
        thisPair = ""
        i = 0
        while i < len(strInput):
            thisPair = strInput[i]
            i += 1
            if i < len(strInput):
                thisPair += strInput[i]
            else:  # pragma: no cover
                thisPair += spaceChar
            strOutput += dHex(thisPair)
            i += 1
        self.state = strOutput.encode()
        return self

    @ChepyDecorators.call_stack
    def from_base36(
        self, delimiter: Union[str, bytes] = " ", join_by: Union[str, bytes] = " "
    ) -> DataFormatT:
        """Decode Base36 data

        Args:
            delimiter (Union[str, bytes], optional): Delimiter to split groups of ints by. Defaults to ' '.
            join_by (Union[str, bytes], optional): Join final output by. Defaults to ' '.

        Returns:
            Chepy: The Chepy object.
        """
        delimiter = self._bytes_to_str(delimiter)
        join_by = self._bytes_to_str(join_by)
        data = self._convert_to_str().split(delimiter)
        hold = []

        alphabet = "0123456789abcdefghijklmnopqrstuvwxyz"
        for number in data:
            number = int(number)
            if number == 0:
                return "0"  # pragma: no cover
            base36 = ""
            while number != 0:
                number, i = divmod(number, 36)
                base36 = alphabet[i] + base36
            hold.append(base36)
        self.state = join_by.join(hold)
        return self

    @ChepyDecorators.call_stack
    def to_base36(self, join_by: Union[str, bytes] = b" ") -> DataFormatT:
        """Encode to Base 36

        Args:
            join_by (Union[str, bytes], optional): Join final output by. Defaults to b' '.

        Returns:
            Chepy: The Chepy object.
        """
        join_by = self._str_to_bytes(join_by)
        data = self._convert_to_str()
        data = re.compile(r"[^a-zA-Z0-9]").sub(" ", data).split()
        hold = []
        for d in data:
            hold.append(str(int(d.strip(), 36)).encode())
        self.state = join_by.join(hold)
        return self

    @ChepyDecorators.call_stack
    def to_pickle(self) -> DataFormatT:
        """Pickle serialize state

        Returns:
            Chepy: The Chepy object.
        """
        self.state = pickle.dumps(self.state)
        return self

    @ChepyDecorators.call_stack
    def from_pickle(self, trust: bool = False) -> DataFormatT:
        """Deserialize pickle data

        Args:
            trust (bool, optional): As this can lead to code execution, this is a safety net and needs to be set to True. Defaults to False.

        Returns:
            Chepy: The Chepy object.
        """
        if not trust:
            return self
        self.state = pickle.loads(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def to_bacon(
        self,
        A: Literal["A", "0"] = "A",
        B: Literal["B", "1"] = "B",
        complete: bool = True,
        join_by: Union[str, bytes] = b" ",
        invert: bool = False,
    ) -> DataFormatT:
        """Bacon encode

        Args:
            A (Literal['A', '0'], optional): The A character. Defaults to 'A'.
            B (Literal['B', '1'], optional): The B character. Defaults to 'B'.
            complete (bool, optional): Use unique mapping for all characters. Defaults to True.
            join_by (Union[str,bytes], optional): Join output by. Defaults to b' '.
            invert (bool, optional): Invert encoding. Defaults to False.

        Returns:
            Chepy: The Chepy object.
        """
        join_by = self._str_to_bytes(join_by)
        hold = []
        for s in self._convert_to_str():
            if s in string.ascii_letters:
                if complete:
                    hold.append(Encoding.BACON_26.get(s.upper()))
                else:  # pragma: no cover
                    hold.append(Encoding.BACON_24.get(s.upper()))
        updated = []
        for h in hold:
            if invert:  # pragma: no cover
                updated.append(h.replace("a", B).replace("b", A).encode())
            else:
                updated.append(h.replace("a", A).replace("b", B).encode())
        self.state = join_by.join(updated)
        return self

    @ChepyDecorators.call_stack
    def from_bacon(
        self,
        A: Literal["A", "0"] = "A",
        B: Literal["B", "1"] = "B",
        complete: bool = True,
        split_by: Union[str, bytes] = b" ",
        invert: bool = False,
    ) -> DataFormatT:
        """From Bacon

        Args:
            A (Literal['A','0'], optional): A character. Defaults to 'A'.
            B (str, optional): B character. Defaults to 'B'.
            complete (bool, optional): Use unique mapping for all characters. Defaults to True.
            split_by (Union[str,bytes], optional): Split by. Defaults to b' '.
            invert (bool, optional): Invert decoding. Defaults to False.

        Returns:
            Chepy: The Chepy object.
        """
        split_by = self._bytes_to_str(split_by)
        if complete:
            mapping = {v: k for k, v in Encoding.BACON_26.items()}
        else:
            mapping = {v: k for k, v in Encoding.BACON_26.items()}  # pragma: no cover
        data = self.state
        if not isinstance(self.state, list):  # pragma: no cover
            data = self._convert_to_str().split(split_by)
        out = ""
        for d in data:
            if invert:  # pragma: no cover
                d = d.replace(A, "b").replace(B, "a")
            else:
                d = d.replace(A, "a").replace(B, "b")
            out += mapping.get(d, "")
        self.state = out.encode()
        return self

    @ChepyDecorators.call_stack
    def to_upside_down(self, reverse: bool = False):
        """To upside down

        Args:
            reverse (bool, optional): Reverse order. Defaults to False.

        Returns:
            Chepy: The Chepy object.
        """
        hold = ""
        for s in self._convert_to_str():
            hold += Encoding.UPSIDE_DOWN.get(s, s)
        if reverse:
            self.state = hold[::-1]
        else:
            self.state = hold
        return self

    @ChepyDecorators.call_stack
    def from_upside_down(self, reverse: bool = False):
        """From upside down

        Args:
            reverse (bool, optional): Reverse order. Defaults to False.

        Returns:
            Chepy: The Chepy object.
        """
        encoding = {v: k for k, v in Encoding.UPSIDE_DOWN.items()}
        hold = ""
        for s in self._convert_to_str():
            hold += encoding.get(s, s)
        if reverse:
            self.state = hold[::-1]
        else:
            self.state = hold
        return self

    @ChepyDecorators.call_stack
    def to_messagepack(self) -> DataFormatT:
        """To MessagePack

        Returns:
            Chepy: The Chepy object.
        """
        self.state = msgpack.packb(self.state)
        return self

    @ChepyDecorators.call_stack
    def from_messagepack(self) -> DataFormatT:
        """From MessagePack

        Returns:
            Chepy: The Chepy object.
        """
        self.state = msgpack.unpackb(self.state, raw=False)
        return self

    @ChepyDecorators.call_stack
    def unicode_escape(
        self, padding: int = 0, uppercase_hex: bool = False
    ) -> DataFormatT:
        """Unicode escape

        Args:
            padding (int, optional): Optional padding. Defaults to 0.
            uppercase_hex (bool, optional): Uppercase hex chars. Defaults to False.

        Returns:
            Chepy: The Chepy object.
        """

        def unicode_replacer(match):
            code_point = ord(match.group(0))
            padding_format = "{:04x}".format(code_point)
            if uppercase_hex:
                padding_format = padding_format.upper()
            return r"\u" + "0" * padding + padding_format

        escaped_string = re.sub(
            r"[^\x00-\x7F]", unicode_replacer, self._convert_to_str()
        )
        self.state = escaped_string
        return self

    @ChepyDecorators.call_stack
    def to_base(self, radix: int = 36) -> DataFormatT:
        """Convert int to base

        Args:
            radix (int, optional): Radix. Defaults to 36.

        Returns:
            Chepy: The Chepy object.
        """
        num = self._convert_to_int()
        if num == 0:
            self.state = "0"
            return self

        chars = string.digits + string.ascii_lowercase
        result = ""

        while num > 0:
            remainder = num % radix
            result = chars[remainder] + result
            num //= radix

        self.state = result
        return self

    @ChepyDecorators.call_stack
    def from_base(self, radix: int = 36) -> DataFormatT:
        """Convert string to int base

        Args:
            radix (int, optional): Radix. Defaults to 36.

        Returns:
            Chepy: The Chepy object.
        """
        chars = string.digits + string.ascii_lowercase
        result = 0

        string_num = self._convert_to_str()
        for char in string_num:
            result = result * radix + chars.index(char)

        self.state = result
        return self

    @ChepyDecorators.call_stack
    def rotate_right(self, radix: int = 1, carry: bool = False) -> DataFormatT:
        """Rotate right

        Args:
            radix (int, optional): Radix. Defaults to 1.
            carry (bool, optional): Carry. Defaults to False.

        Returns:
            Chepy: The Chepy object.
        """
        r = Rotate(self._convert_to_bytes(), radix)
        if carry:
            self.state = r.rot_right_carry()
        else:
            self.state = r.rot(Rotate.rotate_right)
        return self

    @ChepyDecorators.call_stack
    def rotate_left(self, radix: int = 1, carry: bool = False) -> DataFormatT:
        """Rotate left

        Args:
            radix (int, optional): Radix. Defaults to 1.
            carry (bool, optional): Carry. Defaults to False.

        Returns:
            Chepy: The Chepy object.
        """
        r = Rotate(self._convert_to_bytes(), radix)
        if carry:
            self.state = r.rotate_left_carry()
        else:
            self.state = r.rot(Rotate.rotate_left)
        return self

    @ChepyDecorators.call_stack
    def to_base62(
        self,
        alphabet: str = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz",
    ) -> DataFormatT:
        """Encode to base62

        Args:
            alphabet (str, optional): Alphabet. Defaults to "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz".

        Returns:
            Chepy: The Chepy object
        """
        base62 = []
        num = int.from_bytes(self._convert_to_bytes(), byteorder="big")

        while num > 0:
            num, remainder = divmod(num, 62)
            base62.insert(0, alphabet[remainder])

        self.state = "".join(base62)
        return self

    @ChepyDecorators.call_stack
    def from_base65536(self) -> DataFormatT:
        """Decode Base 65536 for unicode

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_str()
        stream = io.BytesIO()
        done = False
        for ch in data:
            code_point = ord(ch)
            b1 = code_point & ((1 << 8) - 1)
            try:
                b2 = Base65536.B2[code_point - b1]
            except KeyError:  # pragma: no cover
                raise ValueError("Invalid base65536 code point: %d" % code_point)
            b = int2byte(b1) if b2 == -1 else int2byte(b1) + int2byte(b2)
            if len(b) == 1:
                if done:  # pragma: no cover
                    raise ValueError("base65536 sequence continued after final byte")
                done = True
            stream.write(b)
        self.state = stream.getvalue()
        return self

    @ChepyDecorators.call_stack
    def to_base65536(self) -> DataFormatT:
        """Encode to Base 65536. Base64 for unicode

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_bytes()
        stream = io.StringIO()
        length = len(data)
        for x in range(0, length, 2):
            b1 = indexbytes(data, x)
            b2 = indexbytes(data, x + 1) if x + 1 < length else -1
            code_point = Base65536.BLOCK_START[b2] + b1
            stream.write(unichr(code_point))
        self.state = stream.getvalue()
        return self

    @ChepyDecorators.call_stack
    def from_base62(
        self,
        alphabet: str = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz",
    ) -> DataFormatT:
        """Decode from base62

        Args:
            alphabet (str, optional): Alphabet. Defaults to "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz".

        Returns:
            Chepy: The Chepy object.
        """
        base62_dict = {char: index for index, char in enumerate(alphabet)}
        num = 0
        for char in self._convert_to_str():
            num = num * 62 + base62_dict[char]
        decoded_data = num.to_bytes((num.bit_length() + 7) // 8, byteorder="big")

        self.state = decoded_data
        return self

    @ChepyDecorators.call_stack
    def cut(self, start: int, end: int) -> DataFormatT:
        """Convert the state to bytes and cut x:y data from it

        Args:
            start (int): Starting position
            end (int): End position

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_bytes()
        self.state = data[start:end]
        return self

    @ChepyDecorators.call_stack
    def flatten(self) -> DataFormatT:
        """Flatten a list of lists into a single list

        Returns:
            Chepy: The Chepy object.
        """
        try:
            self.state = list(itertools.chain(*self.state))
            return self
        except:
            return self

    @ChepyDecorators.call_stack
    def to_utf21(self) -> DataFormatT:
        """Convert to UTF-21

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_bytes()
        bits = Uint1Array(len(data) * 21)

        bit_index = 0
        for codepoint in data:
            for i in range(20, -1, -1):
                bits.set(bit_index, (codepoint & (1 << i)) >> i)
                bit_index += 1

        self.state = bytes(bits.buffer)
        return self

    @ChepyDecorators.call_stack
    def from_utf21(self) -> DataFormatT:
        """Convert from UTF-21

        Returns:
            Chepy: The Chepy object.
        """
        buffer = self._convert_to_bytes()
        bits = Uint1Array(buffer)
        codepoint_length = (len(buffer) * 8) // 21
        codepoints = []

        for codepoint_index in range(codepoint_length):
            start_bit_index = codepoint_index * 21
            codepoint = 0

            for i in range(21):
                bit = bits.get(start_bit_index + i)
                codepoint += bit << (20 - i)

            codepoints.append(codepoint)

        self.state = bytes(codepoints)
        return self

    @ChepyDecorators.call_stack
    def to_uuencode(self, header: str = "-") -> DataFormatT:
        """To UUEncode

        Args:
            header (str): header

        Returns:
            Chepy: The Chepy object.
        """
        self.state = UUEncoderDecoder(self._convert_to_bytes(), header).uuencode()
        return self

    @ChepyDecorators.call_stack
    def from_uuencode(self: DataFormatT, header: str = "-") -> DataFormatT:
        """From UUEncode

        Args:
            header (str): header

        Returns:
            Chepy: The Chepy object.
        """
        self.state = UUEncoderDecoder(self._convert_to_bytes(), header).uudecode()
        return self

    @ChepyDecorators.call_stack
    def from_quoted_printable(self) -> DataFormatT:
        """From quoted printable

        Returns:
            Chepy: The Chepy object.
        """
        self.state = quopri.decodestring(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def to_quoted_printable(self) -> DataFormatT:
        """To quoted printable

        Returns:
            Chepy: The Chepy object.
        """
        self.state = quopri.encodestring(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def from_rison(self) -> DataFormatT:
        """Encode to RISON

        Returns:
            Chepy: The Chepy object.
        """
        self.state = rison.loads(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def to_rison(self) -> DataFormatT:
        """Decode from RISON

        Returns:
            Chepy: The Chepy object.
        """
        self.state = rison.dumps(self.state)
        # if option is None:
        # elif option == 'array':
        #     self.state = rison.encode_array(self.state)
        # elif option == 'object':
        #     self.state = rison.encode_object(self.state)
        # elif option == 'uri':
        #     self.state = rison.encode_uri(self.state)
        # else:
        #     self._log.error('Invalid data type')
        return self

    @ChepyDecorators.call_stack
    def increment_bytes(self, n: int) -> DataFormatT:
        """Loop through each byte and increment

        Args:
            n (int): increment by.

        Returns:
            Chepy: The Chepy object.
        """
        count = int(n)
        hold = bytearray()
        data = self._convert_to_bytes()
        for d in data:
            hold.append(d + count)
        self.state = bytes(hold)
        return self

    @ChepyDecorators.call_stack
    def decrement_bytes(self, n: int) -> DataFormatT:
        """Loop through each byte and decrement

        Args:
            n (int): decrement by.

        Returns:
            Chepy: The Chepy object.
        """
        count = int(n)
        hold = bytearray()
        data = self._convert_to_bytes()
        for d in data:
            hold.append(d - count)
        self.state = bytes(hold)
        return self

    @ChepyDecorators.call_stack
    def parse_csv(self) -> DataFormatT:
        """Parse a csv file. Returns a list of dict objects.

        Returns:
            Chepy: The Chepy object.
        """
        data = io.StringIO(self._convert_to_str())
        rows = csv.DictReader(data)
        self.state = [x for x in rows]
        return self

    @ChepyDecorators.call_stack
    def parse_sqlite(self, query: str) -> DataFormatT:
        """Parse sqlite db and run queries against it. Returns an array of dict objects with column as key and value

        Args:
            query (str): SQL Query

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_bytes()
        query = self._bytes_to_str(query)
        conn = sqlite3.connect(":memory:")
        conn.deserialize(data)
        cursor = conn.cursor()

        cursor.execute(query)

        rows = cursor.fetchall()
        hold = []
        columns = [col[0] for col in cursor.description]

        for r in rows:
            hold.append(dict(collections.OrderedDict(zip(columns, r))))

        self.state = hold
        return self

    @ChepyDecorators.call_stack
    def to_italics(self) -> DataFormatT:
        """Convert string alphabets to italics

        Returns:
            Chepy: The Chepy object.
        """
        text = self._convert_to_str()
        normal = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
        italic = "𝘢𝘣𝘤𝘥𝘦𝘧𝘨𝘩𝘪𝘫𝘬𝘭𝘮𝘯𝘰𝘱𝘲𝘳𝘴𝘵𝘶𝘷𝘸𝘹𝘺𝘻𝘈𝘉𝘊𝘋𝘌𝘍𝘎𝘏𝘐𝘑𝘒𝘓𝘔𝘕𝘖𝘗𝘘𝘙𝘚𝘛𝘜𝘝𝘞𝘟𝘠𝘡"
        translation_table = str.maketrans(normal, italic)
        self.state = text.translate(translation_table)
        return self

    @ChepyDecorators.call_stack
    def from_italics(self) -> DataFormatT:
        """Convert italic string alphabets to ascii

        Returns:
            Chepy: The Chepy object.
        """
        text = self._convert_to_str()
        normal = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
        italic = "𝘢𝘣𝘤𝘥𝘦𝘧𝘨𝘩𝘪𝘫𝘬𝘭𝘮𝘯𝘰𝘱𝘲𝘳𝘴𝘵𝘶𝘷𝘸𝘹𝘺𝘻𝘈𝘉𝘊𝘋𝘌𝘍𝘎𝘏𝘐𝘑𝘒𝘓𝘔𝘕𝘖𝘗𝘘𝘙𝘚𝘛𝘜𝘝𝘞𝘟𝘠𝘡"
        translation_table = str.maketrans(italic, normal)
        self.state = text.translate(translation_table)
        return self

    @ChepyDecorators.call_stack
    def to_fullwidth(self):
        """Convert string to full width string

        Returns:
            Chepy: The Chepy object.
        """
        text = self._convert_to_str()
        result = ""
        for char in text:
            if ord(char) in range(33, 127):
                result += chr(ord(char) - 33 + 0xFF01)
            elif char == " ":
                result += "\u3000"
            else:
                result += char
        self.state = result
        return self

    @ChepyDecorators.call_stack
    def from_fullwidth(self):
        """Convert full width string to regular string

        Returns:
            Chepy: The Chepy object.
        """
        text = self._convert_to_str()
        result = ""
        for char in text:
            if ord(char) in range(0xFF01, 0xFF5F):  # Fullwidth ASCII range
                result += chr(ord(char) - 0xFF01 + 33)
            elif char == "\u3000":  # Ideographic space
                result += " "
            else:
                result += char  # pragma: no cover
        self.state = result
        return self
